import copy
import time

from ruamel.yaml.comments import CommentedSeq
from ruamel.yaml.representer import RoundTripRepresenter

from . import seq

import stickydesign as sd
import stickydesign.stickydesign2 as sd2

DEFAULT_ENERGETICS = sd.EnergeticsDAOE(temperature=33, coaxparams=True)

DEFAULT_MULTIMODEL_ENERGETICS = [
    sd.EnergeticsDAOE(temperature=33, coaxparams='protozanova'),
    sd.EnergeticsDAOE(temperature=33, coaxparams='pyshni'),
    sd.EnergeticsDAOE(temperature=33, coaxparams='peyret'),
    sd.EnergeticsDAOE(temperature=33, coaxparams=False)
]

DEFAULT_SD2_MULTIMODEL_ENERGETICS = [
    sd2.EnergeticsDAOEC(5, temperature=36, coaxparams='protozanova'),
    sd2.EnergeticsDAOEC(5, temperature=36, coaxparams='pyshni'),
    sd2.EnergeticsDAOEC(5, temperature=36, coaxparams='peyret'),
    sd2.EnergeticsDAOEC(5, temperature=36, coaxparams=False)
]

DEFAULT_MM_ENERGETICS_NAMES = ['Prot', 'Pysh', 'Peyr', 'None']

DEFAULT_REGION_ENERGETICS = sd.EnergeticsBasic(
    temperature=33, coaxparams=False, danglecorr=False)

MAPPER = map

MPOBJECT = None


def multimap(function, data):
    return MAPPER(function, data)


def setup_multi(method, ncores=None):
    global MPOBJECT
    global MAPPER
    if method == 'none':
        MAPPER = map
    elif method == 'multiprocessing':
        import multiprocessing
        if not ncores:
            import os
            ncores = os.cpu_count() - 1
        if not isinstance(MPOBJECT, multiprocessing.pool.Pool):
            MPOBJECT = multiprocessing.Pool(ncores)
        MAPPER = MPOBJECT.map
    elif method == 'ipyparallel':
        import ipyparallel
        if not isinstance(MPOBJECT, ipyparallel.client.view.LoadBalancedView):
            rc = ipyparallel.Client()
            MPOBJECT = rc.load_balanced_view()
        MAPPER = MPOBJECT.map


class NamedList(CommentedSeq):
    """A class for a list of dicts, where some dicts have a 'name' item
which should be unique, but others might not.  Indexing works with
either number or name.  Note that updating dicts may make the list
inconsistent.

    """

    def __init__(self, x=[]):
        CommentedSeq.__init__(self, x)

    def __getitem__(self, i):
        if isinstance(i, str):
            r = [x for x in self if x.get('name', None) == i]
            if len(r) > 1:
                raise KeyError(
                    "There are {} elements named {}.".format(len(r), i))
            elif len(r) == 0:
                raise KeyError("No element named {} found.".format(i))
            else:
                return r[0]
        else:
            return CommentedSeq.__getitem__(self, i)

    def check_consistent(self):
        """Checks that each name appears only once.  On failure, returns a
ValueError with (message, {failed_name: count}).  Otherwise, return
with no output.

        """
        names = [v['name'] for v in self if 'name' in v.keys()]

        # if we have *no* names, the tests will fail, but we are obviously
        # consistent, so:
        if not names:
            return

        from collections import Counter
        namecounts = Counter(names)

        if max(namecounts.values()) > 1:
            badcounts = {n: v for n, v in namecounts.items() if v > 1}
            raise ValueError("Inconsistent NamedList.", badcounts)

    def __setitem__(self, i, v):
        if isinstance(i, str):
            r = [(ii, x) for ii, x in enumerate(self)
                 if x.get('name', None) == i]
            if len(r) > 1:
                raise KeyError(
                    "There are {} elements named {}.".format(len(r), i))
            elif len(r) == 0:
                self.append(v)
            else:
                CommentedSeq.__setitem__(self, r[0][0], v)
        else:
            CommentedSeq.__setitem__(self, i, v)

    def __delitem__(self, i):
        if isinstance(i, str):
            r = [ii for ii, x in enumerate(self) if x.get('name', None) == i]
            if len(r) > 1:
                raise KeyError(
                    "There are {} elements named {}.".format(len(r), i))
            elif len(r) == 0:
                raise KeyError("No element named {} found.".format(i))
            else:
                CommentedSeq.__delitem__(self, r[0])
        else:
            return CommentedSeq.__delitem__(self, i)

    def keys(self):
        return [x['name'] for x in self if 'name' in x.keys()]

    def __deepcopy__(self, memo):
        # FIXME: this is here waiting on ruamel.yaml bugfix.
        res = self.__class__()
        memo[id(self)] = res
        for k in self:
            res.append(copy.deepcopy(k))
            self.copy_attributes(res, deep=True)
        return res


RoundTripRepresenter.add_representer(NamedList,
                                     RoundTripRepresenter.represent_list)

import numpy as np
RoundTripRepresenter.add_representer(np.str_,
                                     RoundTripRepresenter.represent_str)


class ProgressLogger(object):
    def __init__(self, logger, N, seconds_interval=60):
        self.logger = logger
        stime = time.perf_counter()
        self.stime = stime
        self.ltime = stime
        self.li = 0
        self
        self.seconds_interval = seconds_interval
        self.N = N
        self.logger.info("starting {} tasks".format(self.N))

    def update(self, i):
        ctime = time.perf_counter()
        if ctime - self.ltime > self.seconds_interval:
            self.logger.info(
                "finished {}/{}, {} s elapsed, {} s est remaining".format(
                    i, self.N,
                    int(ctime - self.stime),
                    int((self.N - i) * (ctime - self.stime) / i)))
            self.ltime = ctime
            self.li = i


def comp(endname):
    "Return the complementary name of a given end (eg, for 'a', return 'a/')"
    if endname[-1] == '/':
        return endname[:-1]
    else:
        return endname + '/'


def base(endname):
    "Return the base name of a given end name (eg, for either 'a' or 'a/', return 'a')"
    if endname[-1] == '/':
        return endname[:-1]
    else:
        return endname


class GlueMergeSpec:
    def __init__(self, ops=[]):
        self._cm = {}
        self._ecs = []
        for op in ops:
            self.add(*op)

    def add(self, a, b):
        ai = self._cm.get(a, -1)
        bi = self._cm.get(b, -1)
        if (ai > -1) and (bi == -1):
            self._ecs[ai].add(b)
            self._ecs[self._cm[comp(a)]].add(comp(b))
            self._rebuild_map()
        if (ai == -1) and (bi > -1):
            self._ecs[bi].add(a)
            self._ecs[self._cm[comp(b)]].add(comp(a))
            self._rebuild_map()
        if (ai == -1) and (bi == -1):
            self._ecs.append({a, b})
            self._ecs.append({comp(a), comp(b)})
            self._rebuild_map()
        if (ai > -1) and (bi > -1):
            if ai == bi:
                return
            if ai == self._cm[comp(b)]:
                raise ValueError
            self._ecs[ai].update(self._ecs[bi])
            self._ecs[self._cm[comp(a)]].update(self._ecs[self._cm[comp(b)]])
            del (self._ecs[bi])
            if self._cm[comp(b)] > bi:
                del (self._ecs[self._cm[comp(b)] - 1])
            else:
                del (self._ecs[self._cm[comp(b)]])
            self._rebuild_map()

    def copyadd(self, a, b):
        v = copy.deepcopy(self)
        v.add(a, b)
        return v

    def _rebuild_map(self):
        for i, s in enumerate(self._ecs):
            for v in s:
                if comp(v) in s:
                    raise ValueError
                self._cm[v] = i

    def eq(self, a, b):
        if a == b:
            return True
        elif (a in self._cm.keys()) and (b in self._cm.keys()) and (
                self._cm[a] == self._cm[b]):
            return True
        else:
            return False


class TileMergeSpec:
    def __init__(self, ops=[]):
        self._cm = {}
        self._ecs = []
        for op in ops:
            self.add(*op)

    def add(self, a, b):
        ai = self._cm.get(a, -1)
        bi = self._cm.get(b, -1)
        if (ai > -1) and (bi == -1):
            self._ecs[ai].add(b)
            self._rebuild_map()
        if (ai == -1) and (bi > -1):
            self._ecs[bi].add(a)
            self._rebuild_map()
        if (ai == -1) and (bi == -1):
            self._ecs.append({a, b})
            self._rebuild_map()
        if (ai > -1) and (bi > -1):
            if ai == bi:
                return
            self._ecs[ai].update(self._ecs[bi])
            del (self._ecs[bi])
            self._rebuild_map()

    def copyadd(self, a, b):
        v = copy.deepcopy(self)
        v.add(a, b)
        return v

    def _rebuild_map(self):
        for i, s in enumerate(self._ecs):
            for v in s:
                self._cm[v] = i

    def eq(self, a, b):
        if a == b:
            return True
        elif (a in self._cm.keys()) and (b in self._cm.keys()) and (
                self._cm[a] == self._cm[b]):
            return True
        else:
            return False



import itertools
from .util import comp, GlueMergeSpec
import numpy as np
import re
from .tiles import TileList
from .sensitivitynew import _fakesingles

_OE = np.array([2, 3, 0, 1])
_ENDER = np.array([3, 2, 1, 0])


def _branchtiles(gl, st, tns, direction, equiv=None, onlyusedandfs=True):
    """For an array of tile number chains (newest-to-oldest along dim 1) 
       in a FTilesArray, find the tiles that match the direction (with equiv).
       Return a new chain array with the new tiles added."""
    if tns is None:
        return None
    newtnl = []
    for tn in tns:
        sel = (equiv[gl.complement[st.glues[tn[0], direction]]] ==
               equiv[st.glues[:, _OE[direction]]])
        if onlyusedandfs:
            sel = sel & (st.used)
        ntns = np.flatnonzero(sel)
        ftns = np.empty((len(ntns), len(tn)+1), dtype='int')
        ftns[:, 0] = ntns
        ftns[:, 1:] = tn[None, :]
        newtnl.append(ftns)
    if len(newtnl) == 0:
        return None
    return np.concatenate(newtnl, axis=0)


def _latticedefect_tile(gl, st, tn, direction='e', n=2,
                        equiv=None):
    """With n tiles in each branch, can tile number tn form
    a lattice defect?"""
    if equiv is None:
        equiv = gl.blankequiv()
    d1, d2 = {'e': (1, 2), 'w': (0, 3), 'n': (0, 1), 's': (2, 3)}[direction]
    b1 = np.array([[tn]], dtype='int')
    b2 = b1
    for _ in range(0, n):
        b1 = _branchtiles(gl, st, b1, d1, equiv=equiv)
        b2 = _branchtiles(gl, st, b2, d2, equiv=equiv)

    if (b1 is None) or (b2 is None):
        return None
    tiledirs = _OE[_ENDER[[d1, d2]]]
    # shape here is: 
    tilegluecomps = equiv[gl.complement[st.glues[:, tiledirs]]]

    c1g = st.glues[b1[:, 0], _ENDER[d1]]
    c2g = st.glues[b2[:, 0], _ENDER[d2]]

    # dims here become (CLOSER, BRANCH)
    s1 = tilegluecomps[:, 0, None] == c1g[None, :]
    s2 = tilegluecomps[:, 1, None] == c2g[None, :]

    # dims here become (CLOSER, BRANCH1, BRANCH2)
    ts = (s1[:, :, None] & s2[:, None, :]) & (st.used)[:, None, None]
    
    r = np.nonzero(ts)

    if len(r[0]) == 0:
        return None
    else:
        return r

    
def latticedefects(fts,
                   direction='e',
                   depth=2,
                   rotate=False,
                   equiv=None):
    if equiv is None:
        equiv = fts.blankequiv()
    st = fts.tilelist.stiles
    gl = fts.gluelist
    
    itns = np.flatnonzero(st.used)

    ldl = []
    for tn in itns:
        r = _latticedefect_tile(gl, st, tn, direction, n=depth, equiv=equiv)
        if r is None:
            continue
        ldl.append((tn, r))
    return ldl



import ruamel.yaml as yaml
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.representer import RoundTripRepresenter
import warnings

import copy
import re

import pkg_resources
import os

from .tiles import TileList
from .ends import EndList

from . import tilestructures
from . import seeds
from . import util
from . import seq
from . import sensitivitynew as sensitivity
from . import fastreduce

from peppercompiler import compiler as compiler
from peppercompiler.design import spurious_design as spurious_design
from peppercompiler import finish as finish
from peppercompiler.DNA_classes import wc

import numpy as np
import stickydesign as sd

import stickydesign.multimodel as multimodel
from collections import Counter

import collections
from random import shuffle
from datetime import datetime, timezone

import logging

import stickydesign.stickydesign2 as sd2

from .util import DEFAULT_SD2_MULTIMODEL_ENERGETICS, DEFAULT_MM_ENERGETICS_NAMES, DEFAULT_REGION_ENERGETICS, DEFAULT_MULTIMODEL_ENERGETICS, DEFAULT_ENERGETICS

SELOGGER = logging.getLogger(__name__)


class TileSet(CommentedMap):
    """A class representing a single DNA tile assembly system.

    This class parses and manipulates Alhambra tilesets.  Note that,
    unless you have already imported or made a dict, you will likely
    want to construct this with TileSet.from_file .

    Parameters
    ----------

    val : dict
        a dict in the Alhambra tileset format.

    Methods
    -------
    
    from_file(name_or_stream)
        (class method) create a TileSet from a filename or stream.
    """

    def __init__(self, val={}):
        CommentedMap.__init__(self, val)

        self['ends'] = EndList(self.get('ends', []))
        self['tiles'] = TileList(self.get('tiles', []))

    @classmethod
    def from_file(cls, name_or_stream, *args, **kwargs):
        """
        Class method to create a TileSet from a file or stream.

        Parameters
        ----------
        name_or_stream : file-like or path-like object
            The path of the file, or a file object, to lead.


        Returns
        -------
        TileSet
            The loaded TileSet.

        """
        # Assume a stream:
        if getattr(name_or_stream, 'read', None) is None:
            return cls(
                yaml.round_trip_load(
                    open(name_or_stream, 'r'), *args, **kwargs))
        else:
            return cls(yaml.round_trip_load(name_or_stream, *args, **kwargs))

    def to_file(self, name_or_stream):
        """
        Write a TileSet to a file or stream.

        Parameters
        ----------
        name_or_stream : file-like or path-like object
            The path of the file, or a file object, to write to.

        """

        if getattr(name_or_stream, 'write', None) is None:
            return yaml.round_trip_dump(self, open(name_or_stream, 'w'))
        else:
            return yaml.round_trip_dump(self, name_or_stream)

    @property
    def tiles(self):
        """The TileList of tiles in the TileSet

        Returns
        -------

        TileList"""
        return self['tiles']

    @tiles.setter
    def tiles(self, val):
        if isinstance(val, TileList):
            self['tiles'] = val
        else:
            self['tiles'] = TileList(val)

    @property
    def allends(self):
        """All ends in the system, both from ends and tiles.

        Returns
        -------

        EndList

        See Also
        --------

        TileSet.ends
        TileSet.tiles.endlist"""
        return self.ends.merge(self.tiles.endlist())

    def ends():
        doc = """The EndList of specified ends in the TileSet (not including
                 ends that are only in tiles.

                 Returns
                 -------

                 EndList

                 See Also
                 --------

                 TileSet.allends

                 TileSet.tiles.endlist"""

        def fget(self):
            return self['ends']

        def fset(self, value):
            self['ends'] = value

        def fdel(self):
            del self['ends']

        return locals()

    ends = property(**ends())

    @property
    def seed(self):
        return self.get('seed', None)

    def create_abstract_diagrams(self, filename, *options):
        """Write abstract tile diagrams in SVG for the TileSet.

        Parameters
        ----------

        filename : str
            a path or filename for the SVG output.
        *options
            (currently unused)
        """
        from lxml import etree
        import os
        import pkg_resources
        base = etree.parse(
            pkg_resources.resource_stream(
                __name__, os.path.join('seqdiagrambases', 'blank.svg')))
        baseroot = base.getroot()

        pos = 0
        for tile in self.tiles:
            group, n = tile.abstract_diagram(self)

            group.attrib['transform'] = "translate({},{})".format(
                (pos % 12) * 22, (pos // 12) * 22)
            pos += n

            baseroot.append(group)

        base.write(filename)

    def create_sequence_diagrams(tileset, filename, *options):
        """Write sequence tile diagrams in SVG for the TileSet.

        Parameters
        ----------

        filename : str
            a path or filename for the SVG output.
        *options
            (currently unused)
        """

        from lxml import etree
        import pkg_resources
        import os.path

        base = etree.parse(
            pkg_resources.resource_stream(
                __name__, os.path.join('seqdiagrambases', 'blank.svg')))
        baseroot = base.getroot()
        pos = 150
        for tile in tileset.tiles:
            group = tile.sequence_diagram()

            group.attrib['transform'] = 'translate(0,{})'.format(pos)
            pos += 150
            baseroot.append(group)

        base.write(filename)

    def create_adapter_sequences(tileset):
        """Generate adapter sequences for the TileSet, returning a new
        TileSet with the generated sequences included.

        Returns
        -------

        TileSet
           TileSet with generated adapter sequences.
        """
        seedclass = seeds.seedtypes[tileset['seed']['type']]
        if seedclass.needspepper:
            warnings.warn("This set must have adapter sequences created during\
     regular sequence design. You can ignore this if you just created sequences."
                          )
            return tileset
        return seedclass.create_adapter_sequences(tileset)

    def create_layout_diagrams(tileset,
                               xgrowarray,
                               filename,
                               scale=1,
                               *options):
        """Create an SVG layout diagram from xgrow output.

        This currently uses the abstract diagram bases to create the
        layout diagrams.

        Parameters
        ----------

        xgrowarray : ndarray or dict
            Xgrow output.  This may be a numpy array of
            an xgrow state, obtained in some way, or may be the 'array'
            output of xgrow.run.

        filename : string
            File name / path of the output file.

        """
        from lxml import etree
        base = etree.parse(
            pkg_resources.resource_stream(
                __name__, os.path.join('seqdiagrambases', 'blank.svg')))
        baseroot = base.getroot()

        svgtiles = {}

        if isinstance(xgrowarray, dict):
            if 'tiles' in xgrowarray:
                xgrowarray = xgrowarray['tiles']
            elif 'array' in xgrowarray:
                xgrowarray = xgrowarray['array']['tiles']

        for tile in tileset.tiles:
            group, n = tile.abstract_diagram(tileset)
            svgtiles[tile.name] = group

        tilelist = tileset.generate_xgrow_dict(perfect=True)['tiles']
        tilen = [None] + [x['name'] for x in tilelist]
        firstxi = 10000
        firstyi = 10000
        import copy
        for yi in range(0, xgrowarray.shape[0]):
            for xi in range(0, xgrowarray.shape[1]):
                tn = tilen[xgrowarray[yi, xi]]
                if tn and tn[-5:] == '_left':
                    tn = tn[:-5]
                if tn and tn[-7:] == '_bottom':
                    tn = tn[:-7]
                if not (tn in svgtiles.keys()):
                    continue
                if xi < firstxi:
                    firstxi = xi
                if yi < firstyi:
                    firstyi = yi
                st = copy.deepcopy(svgtiles[tn])
                st.attrib['transform'] = 'translate({},{})'.format(
                    xi * 10, yi * 10)
                baseroot.append(st)

        base.write(filename)

    @property
    def strand_order_list(self):
        """Return a list of orderable strand sequences for all tile strands.

        These sequences are generated by the Tile.orderableseqs attribute, which
        should include useful strand names and modified bases for labels.  The
        sequences may thus not be valid ACGT-only sequences (eg, they may contain
        /iBiodT/).

        Returns
        -------

        list
            a list of tuple (strand_name, strand_sequence) for the system.
        """
        return [y for x in self.tiles for y in x.orderableseqs]

    def check_consistent(self):
        """Check the TileSet consistency.

        Check a number of properties of the TileSet for consistency.
        In particular:

           * Each tile must pass Tile.check_consistent()
           * TileSet.ends and TileSet.tiles.endlist() must not contain conflicting
             ends or end sequences.
           * If there is a seed:
               * It must be of an understood type (it must be in seeds.seedtypes)
               * All adapter locations must be valid.
               * The seed must pass its check_consistent and check_sequence.
        """
        # * END LIST The end list itself must be consistent.
        # ** Each end must be of understood type
        # ** Each end must have a valid sequence or no sequence
        # ** There must be no more than one instance of each name
        # ** WARN if there are ends with no namecounts
        # * TILE LIST
        # ** each tile must be of understood type (must parse)
        # ** ends in the tile list must be consistent (must merge)
        # ** there must be no more than one tile with each name
        self.tiles.check_consistent()
        endsfromtiles = self.tiles.endlist()

        # ** WARN if any end that appears does not have a complement used or vice versa
        # ** WARN if there are tiles with no name
        # * TILE + END
        # ** The tile and end lists must merge validly
        # (checks sequences, adjacents, types, complements)
        self.ends.merge(endsfromtiles)

        # ** WARN if tilelist has end references not in ends
        # ** WARN if merge is not equal to the endlist
        # ** WARN if endlist has ends not used in tilelist
        # * ADAPTERS / SEEDS
        if 'seed' in self.keys():
            # ** seeds must be of understood type
            assert self['seed']['type'] in seeds.seedtypes.keys()
            # ** adapter locations must be valid
            sclass = seeds.seedtypes[self['seed']['type']]

            sclass.check_consistent(self)

            sclass.check_sequence(self)
            # ** each adapter must have no sequence or a consistent sequence
            # *** the RH strand must match the associated tile
            # *** the ends in the sequence must match the ends in the endlist
            # *** the LH sequence must be validly binding to both RH and
            #     origami
            # ** each adapter must have valid definition, which means for us:
            # *** if both tile mimic and ends are specified, they must match

    def summary(self):
        """Returns a short summary line about the TileSet"""
        self.check_consistent()
        info = {
            'ntiles':
            len(self.tiles),
            'nrt':
            len([x for x in self.tiles if 'fake' not in x.keys()]),
            'nft':
            len([x for x in self.tiles if 'fake' in x.keys()]),
            'nends':
            len(self.ends),
            'ntends':
            len(self.tiles.endlist()),
            'tns':
            " ".join(x.name for x in self.tiles if x.name),
            'ens':
            " ".join(x.name for x in self.ends if x.name),
            'name':
            " {}".format(self['info']['name']) if
            ('info' in self.keys() and 'name' in self['info'].keys()) else ""
        }
        tun = sum(1 for x in self.tiles if 'name' not in x.keys())
        if tun > 0:
            info['tns'] += " ({} unnamed)".format(tun)
        eun = sum(1 for x in self.ends if 'name' not in x.keys())
        if eun > 0:
            info['ens'] += " ({} unnamed)".format(eun)
        if info['nft'] > 0:
            info['nft'] = " (+ {} fake)".format(info['nft'])
        else:
            info['nft'] = ""
        return "TileSet{name}: {nrt} tiles{nft}, {nends} ends, {ntends} ends in tiles.\nTiles: {tns}\nEnds:  {ens}".format(
            **info)

    def __str__(self):
        return self.summary()

    def copy(self):
        """Return a full (deep) copy of the TileSet"""
        return copy.deepcopy(self)

    def __deepcopy__(self, memo):
        # type: (Any) -> Any
        res = self.__class__()
        memo[id(self)] = res
        for k in self:
            res[k] = copy.deepcopy(self[k])
            self.copy_attributes(res, deep=True)
        return res

    def dump(self, stream):
        """Dump the tileset into a stream in YAML format.

        Parameters
        ----------

        stream : stream
            the stream / file object to dump to.
        """

        return yaml.round_trip_dump(self, stream)

    def design_set(tileset,
                   name='tsd_temp',
                   includes=[
                       pkg_resources.resource_filename(__name__,
                                                       'peppercomps-j1')
                   ],
                   energetics=None,
                   stickyopts={},
                   reorderopts={},
                   coreopts={},
                   keeptemp=False):
        """Helper function to design sets from scratch, calling the numerous parts of
        tilesetdesigner. You may want to use the tilesetdesigner shell script
        instead.

        As with other functions in tilesetdesigner, this should not clobber inputs.

        Parameters
        ----------

        name : str, optional
            Base name for temporary files (default tsd_temp).

        Returns
        -------
        TileSet
            A copy of the TileSet with added sequences.

        """
        if energetics is None:
            energetics = DEFAULT_ENERGETICS

        if hasattr(tileset, 'read'):
            tileset = TileSet.load(tileset)
        else:
            tileset = TileSet(tileset)

        tileset.check_consistent()
        tileset_with_ends_randomorder, new_ends = stickyends.create_sequences(
            tileset, energetics=energetics, **stickyopts)
        tileset_with_ends_ordered = stickyends.reorder(
            tileset_with_ends_randomorder,
            newends=new_ends,
            energetics=energetics,
            **reorderopts)
        tileset_with_strands = create_strand_sequences(
            tileset_with_ends_ordered, name, includes=includes, **coreopts)

        if 'guards' in tileset_with_strands.keys():
            tileset_with_strands = create_guard_strand_sequences(
                tileset_with_strands)

        # FIXME: this is temporary, until we have a better way of deciding.
        if 'createseqs' in tileset_with_strands['seed'].keys():
            tileset_with_strands = create_adapter_sequences(
                tileset_with_strands)

        if not keeptemp:
            os.remove(name + '.fix')
            os.remove(name + '.mfe')
            os.remove(name + '.pil')
            os.remove(name + '.save')
            os.remove(name + '.seqs')
            os.remove(name + '.sys')

        tileset_with_strands.check_consistent()
        return tileset_with_strands

    def create_end_sequences(tileset,
                             method='default',
                             energetics=None,
                             trials=100,
                             devmethod='dev',
                             sdopts={},
                             ecpars={},
                             listends=False):
        """Create sticky end sequences for the TileSet, using stickydesign,
        and returning a new TileSet including the ends.


        Parameters
        ----------
        method: {'default', 'multimodel'} 
            if 'default', use the default, single-model sequence design.  
            If 'multimodel', use multimodel end choice.

        energetics : stickydesign.Energetics
            the energetics instance to use for the design, or list
            of energetics for method='multimodel', in which case the first
            will be the primary.  If None (default), will use
            alhambra.DEFAULT_ENERGETICS, or, if method='multimodel', will use
            alhambra.DEFAULT_MM_ENERGETICS.

        trials : int
            the number of trials to attempt. FIXME

        sdopts : dict
            a dictionary of parameters to pass to stickydesign.easy_ends.

        ecpars : dict
            a dictionary of parameters to pass to the endchooser function
            generator (useful only in limited circumstances).

        listends : bool
            if False, return just the TileSet.  If True, return both the
            TileSet and a list of the names of the ends that were created.

        Returns
        -------
        tileset : TileSet 
            TileSet with designed end sequences included.
        new_ends : list 
            Names of the ends that were designed.

        """
        info = {}
        info['method'] = method
        info['time'] = datetime.now(tz=timezone.utc).isoformat()
        info['sd_version'] = sd.version.__version__

        if not energetics:
            if method == 'multimodel':
                energetics = DEFAULT_MULTIMODEL_ENERGETICS
            else:
                energetics = DEFAULT_ENERGETICS
        if method == 'multimodel' and not isinstance(energetics,
                                                     collections.Iterable):
            raise ValueError("Energetics must be an iterable for multimodel.")
        elif method == 'multimodel':
            all_energetics = energetics
            energetics = all_energetics[0]
            info['energetics'] = [str(e) for e in all_energetics]
            info['trails'] = trials
        elif method == 'default':
            info['energetics'] = str(energetics)

        # Steps for doing this:

        # Create a copy of the tileset.
        newtileset = tileset.copy()

        # Build a list of ends from the endlist in the tileset.  Do this
        # by creating a NamedList, then merging them into it.
        ends = EndList()

        if newtileset.ends:
            ends.merge(newtileset.ends, fail_immediate=False, in_place=True)

        # This is the endlist from the tiles themselves.
        if newtileset.tiles:  # maybe you just want ends?
            # this checks for end/complement usage, and whether any
            # previously-describedends are unused
            # FIXME: implement
            # tilestructures.check_end_usage(newtileset.tiles, ends)

            endlist_from_tiles = newtileset.tiles.endlist()

        ends.merge(endlist_from_tiles, in_place=True)

        # Ensure that if there are any resulting completely-undefined ends, they
        # have their sequences removed.
        #for end in ends:
        #    if end.fseq and set(end.fseq) == {'n'}:
        #        del(end.fseq)

        # Build inputs suitable for stickydesign: lists of old sequences for TD/DT,
        # and numbers of new sequences needed.
        oldDTseqs = [
            end.fseq for end in ends
            if end.etype == 'DT' and seq.is_definite(end.fseq)
        ]
        if oldDTseqs:
            oldDTarray = sd.endarray(oldDTseqs, 'DT')
        oldTDseqs = [
            end.fseq for end in ends
            if end.etype == 'TD' and seq.is_definite(end.fseq)
        ]
        if oldTDseqs:
            oldTDarray = sd.endarray(oldTDseqs, 'TD')

        newTD = [
            end for end in ends
            if end.etype == 'TD' and not seq.is_definite(end.fseq)
        ]
        newDT = [
            end for end in ends
            if end.etype == 'DT' and not seq.is_definite(end.fseq)
        ]

        # Deal with energetics, considering potential old sequences.
        # FIXME: EXPLAIN WHAT THIS ABSTRUSE CODE DOES...
        # TODO: tests needs to test this
        targets = []
        if len(oldDTseqs) == 0 and len(oldTDseqs) == 0:
            targets.append(
                sd.enhist('DT', 5, energetics=energetics)[2]['emedian'])
            targets.append(
                sd.enhist('TD', 5, energetics=energetics)[2]['emedian'])
        if len(oldDTseqs) > 0:
            targets.append(energetics.matching_uniform(oldDTarray))
        if len(oldTDseqs) > 0:
            targets.append(energetics.matching_uniform(oldTDarray))
        targetint = np.average(targets)

        if any(not seq.is_null(end.fseq) for end in newTD):
            TDtemplates = [end.fseq for end in newTD]
        else:
            TDtemplates = None
        if any(not seq.is_null(end.fseq) for end in newDT):
            DTtemplates = [end.fseq for end in newDT]
        else:
            DTtemplates = None

        if method == 'default':
            if TDtemplates or DTtemplates:
                raise NotImplementedError
            # Create new sequences.
            newTDseqs = sd.easyends(
                'TD',
                5,
                number=len(newTD),
                energetics=energetics,
                interaction=targetint,
                **sdopts).tolist()

            newDTseqs = sd.easyends(
                'DT',
                5,
                number=len(newDT),
                energetics=energetics,
                interaction=targetint,
                **sdopts).tolist()

        elif method == 'multimodel':
            SELOGGER.info("starting multimodel sticky end generation " +
                          "of TD ends for {} DT and {} TD ends, {} trials.".
                          format(len(newDT), len(newTD), trials))

            newTDseqs = []
            pl = util.ProgressLogger(SELOGGER, trials * 2)
            presetavail = None
            for i in range(0, trials):
                endchooserTD = multimodel.endchooser(
                    all_energetics,
                    templates=TDtemplates,
                    devmethod=devmethod,
                    **ecpars)

                e, presetavail = sd.easyends(
                    'TD',
                    5,
                    number=len(newTD),
                    oldends=oldTDseqs,
                    energetics=energetics,
                    interaction=targetint,
                    echoose=endchooserTD,
                    _presetavail=presetavail,
                    **sdopts)
                newTDseqs.append(e)
                pl.update(i)

            if oldTDseqs:
                tvals = [[
                    e.matching_uniform(oldTDarray[0:1]) for e in all_energetics
                ] * len(newTDseqs)] * len(newTDseqs)
                SELOGGER.debug(tvals[0])
            else:
                tvals = [[e.matching_uniform(x[0:1]) for e in all_energetics]
                         for x in newTDseqs]

            endchoosersDT = [
                multimodel.endchooser(
                    all_energetics,
                    target_vals=tval,
                    templates=DTtemplates,
                    devmethod=devmethod,
                    **ecpars) for tval in tvals
            ]

            SELOGGER.info("generating corresponding DT ends")
            newDTseqs = []
            presetavail = None

            for i, echoose in enumerate(endchoosersDT):
                e, presetavail = sd.easyends(
                    'DT',
                    5,
                    number=len(newDT),
                    oldends=oldDTseqs,
                    energetics=energetics,
                    interaction=targetint,
                    echoose=echoose,
                    _presetavail=presetavail,
                    **sdopts)
                newDTseqs.append(e)

                pl.update(i + trials)

            arr = [[
                sd.endarray(oldTDseqs + x.tolist(), 'TD'),
                sd.endarray(oldDTseqs + y.tolist(), 'DT')
            ] for x, y in zip(newTDseqs, newDTseqs)]

            scores = [
                multimodel.deviation_score(
                    list(e), all_energetics, devmethod=devmethod) for e in arr
            ]

            sort = np.argsort(scores)

            newTDseqs = newTDseqs[sort[0]].tolist()[len(oldTDseqs):]
            newDTseqs = newDTseqs[sort[0]].tolist()[len(oldDTseqs):]
            info['score'] = float(scores[sort[0]])
            info['maxscore'] = float(scores[sort[-1]])
            info['meanscore'] = float(np.mean(scores))

        # FIXME: move to stickydesign
        assert len(newTDseqs) == len(newTD)
        assert len(newDTseqs) == len(newDT)

        # Shuffle the lists of end sequences, to ensure that they're
        # random order, and that ends used earlier in the set are not
        # always better than those used later. But only shuffle if
        # there were no templates:
        if not TDtemplates:
            shuffle(newTDseqs)
        if not DTtemplates:
            shuffle(newDTseqs)

        # Make sure things are consistent if there are templates:
        if TDtemplates:
            for t, s in zip(TDtemplates, newTDseqs):
                seq.merge(t, s)
        if DTtemplates:
            for t, s in zip(DTtemplates, newDTseqs):
                seq.merge(t, s)

        for end, s in zip(newDT, newDTseqs):
            ends[end.name].fseq = s
        for end, s in zip(newTD, newTDseqs):
            ends[end.name].fseq = s

        ends.check_consistent()

        # Ensure that the old and new sets have consistent end definitions,
        # and that the tile definitions still fit.
        tileset.ends.merge(ends)
        newtileset.tiles.endlist().merge(ends)

        newendnames = [e.name for e in newTD] + [e.name for e in newDT]
        info['newends'] = newendnames

        # Apply new sequences to tile system.
        newtileset.ends = ends
        if 'info' not in newtileset.keys():
            newtileset['info'] = {}
        if 'end_design' not in newtileset['info'].keys():
            newtileset['info']['end_design'] = []
        if isinstance('end_design', dict):  # convert old
            newtileset['info']['end_design'] = [
                newtileset['info']['end_design']
            ]
        newtileset['info']['end_design'].append(info)

        return newtileset, newendnames

    def add_info(self, ititle, data):
        if 'info' not in self.keys():
            self['info'] = {}
        if ititle not in self['info'].keys():
            self['info'][ititle] = []
        self['info'][ititle].append(data)

    def reorder_ends(tileset,
                     newends=[],
                     hightemp=0.1,
                     lowtemp=1e-7,
                     steps=45000,
                     update=1000,
                     energetics=None,
                     fsopts={}):
        """Given a tileset dictionary that includes sticky end sequences, reorder these
        to try to optimize error rates.
        """
        from . import endreorder
        from . import anneal

        if energetics is None:
            energetics = DEFAULT_ENERGETICS

        tset = tileset.copy()

        if 'info' not in tset.keys():
            tset['info'] = {}

        reordersys = endreorder.EndSystemFseq(
            tset, newends, energetics=energetics, **fsopts)

        # FIXME: better parameter control here.
        annealer = anneal.Annealer(reordersys.score, reordersys.mutate)

        newstate = annealer.anneal(reordersys.initstate, hightemp, lowtemp,
                                   steps, update)

        # Now take that new state, and apply it to the new tileset.
        seqs = reordersys.slowseqs(newstate[0])
        for end in tset.ends:
            if end.etype in ['DT', 'TD']:
                eloc = reordersys.enlocs[end['name']]
                end.fseq = seqs[eloc[1]].tolist()[eloc[0]]

        ri = {}

        ri['score'] = float(reordersys.score(newstate[0]))

        tset['info']['reorder'] = ri

        # Ensure that only ends in newends moved: that all others remain mergeable:
        if newends:
            old_ends_from_new_set = EndList(end for end in tset.ends
                                            if end['name'] not in newends)
            tileset.ends.merge(old_ends_from_new_set)

        # Ensure system consistency
        tset.check_consistent()
        return tset

    def _create_end_sequences_sd2(tileset,
                                  method='default',
                                  energetics=None,
                                  trials=100,
                                  devmethod='dev',
                                  sdopts={},
                                  ecpars={},
                                  listends=False):
        """Create sticky end sequences for the TileSet, using stickydesign,
        and returning a new TileSet including the ends. USES STICKYDESIGN2.


        Parameters
        ----------
        method: {'default', 'multimodel'} 
            if 'default', use the default, single-model sequence design.  
            If 'multimodel', use multimodel end choice.

        energetics : stickydesign.Energetics
            the energetics instance to use for the design, or list
            of energetics for method='multimodel', in which case the first
            will be the primary.  If None (default), will use
            alhambra.DEFAULT_ENERGETICS, or, if method='multimodel', will use
            alhambra.DEFAULT_MM_ENERGETICS.

        trials : int
            the number of trials to attempt. FIXME

        sdopts : dict
            a dictionary of parameters to pass to stickydesign.easy_ends.

        ecpars : dict
            a dictionary of parameters to pass to the endchooser function
            generator (useful only in limited circumstances).

        listends : bool
            if False, return just the TileSet.  If True, return both the
            TileSet and a list of the names of the ends that were created.

        Returns
        -------
        tileset : TileSet 
            TileSet with designed end sequences included.
        new_ends : list 
            Names of the ends that were designed.

        """
        info = {}
        info['method'] = method
        info['time'] = datetime.now(tz=timezone.utc).isoformat()
        info['sd_version'] = 'stickydesign2-predev'

        if not energetics:
            if method == 'multimodel':
                energetics = DEFAULT_SD2_MULTIMODEL_ENERGETICS
            else:
                energetics = DEFAULT_SD2_MULTIMODEL_ENERGETICS[0]
        if method == 'multimodel' and not isinstance(energetics,
                                                     collections.Iterable):
            raise ValueError("Energetics must be an iterable for multimodel.")
        elif method == 'multimodel':
            all_energetics = energetics
            energetics = all_energetics[0]
            info['energetics'] = [str(e) for e in all_energetics]
            info['trails'] = trials
        elif method == 'default':
            all_energetics = [energetics]
            info['energetics'] = str(energetics)

        # Steps for doing this:
        print(all_energetics)
        # Create a copy of the tileset.
        newtileset = tileset.copy()

        # Build a list of ends from the endlist in the tileset.  Do this
        # by creating a NamedList, then merging them into it.
        ends = EndList()

        if newtileset.ends:
            ends.merge(newtileset.ends, fail_immediate=False, in_place=True)

        # This is the endlist from the tiles themselves.
        if newtileset.tiles:  # maybe you just want ends?
            # this checks for end/complement usage, and whether any
            # previously-describedends are unused
            # FIXME: implement
            # tilestructures.check_end_usage(newtileset.tiles, ends)

            endlist_from_tiles = newtileset.tiles.endlist()

        ends.merge(endlist_from_tiles, in_place=True)

        # Ensure that if there are any resulting completely-undefined ends, they
        # have their sequences removed.
        #for end in ends:
        #    if end.fseq and set(end.fseq) == {'n'}:
        #        del(end.fseq)

        # Build inputs suitable for stickydesign: lists of old sequences for TD/DT,
        # and numbers of new sequences needed.
        oldDTseqs = [
            end.fseq for end in ends
            if end.etype == 'DT' and seq.is_definite(end.fseq)
        ]
        if oldDTseqs:
            oldDTarray = sd2.EndPairArrayDT(oldDTseqs, 'DT')
        else:
            oldDTarray = None
        oldTDseqs = [
            end.fseq for end in ends
            if end.etype == 'TD' and seq.is_definite(end.fseq)
        ]
        if oldTDseqs:
            oldTDarray = sd2.EndPairArrayTD(oldTDseqs, 'TD')
        else:
            oldTDarray = None

        newTD = [
            end for end in ends
            if end.etype == 'TD' and not seq.is_definite(end.fseq)
        ]
        newDT = [
            end for end in ends
            if end.etype == 'DT' and not seq.is_definite(end.fseq)
        ]

        # If there are old sequences, use those to get an interaction
        # target for the filter:
        interactions = []
        if len(oldDTseqs) > 0:
            interaction.append(np.mean(energetics.gse(oldDTarray)))
        if len(oldTDseqs) > 0:
            interactions.append(np.mean(energetics.gse(oldTDarray)))
        if len(oldDTseqs) == 0 and len(oldTDseqs) == 0:
            interactions.append(sd2.genpre('DT', 5, energetics)[1])
            interactions.append(sd2.genpre('TD', 5, energetics)[1])
        interaction = np.average(interactions)

        availendsDT, _, sfDT = sd2.genpre(
            'DT', 5, energetics, interaction=interaction, oldends=oldDTarray)
        availendsTD, _, sfTD = sd2.genpre(
            'TD', 5, energetics, interaction=interaction, oldends=oldTDarray)

        if any(not seq.is_null(end.fseq) for end in newTD):
            raise NotImplementedError
        if any(not seq.is_null(end.fseq) for end in newDT):
            raise NotImplementedError

        newTDseqs = []
        newDTseqs = []

        rejects = 0
        min_score = 1000.0
        SELOGGER.info("starting sticky end generation " +
                      "of TD ends for {} DT and {} TD ends, {} trials.".format(
                          len(newDT), len(newTD), trials))

        for i in range(0, trials):

            if method == 'default':
                chooser = sd2.BasicChooser(energetics, interaction)

            elif method == 'multimodel':
                chooser = sd2.MultiModelChooser(all_energetics, **ecpars)

            eTD = sd2.easyends(
                'TD',
                5,
                number=len(newTD),
                #oldends=oldTDseqs,
                energetics=energetics,
                interaction=interaction,
                chooser=chooser,
                preavail=availendsTD,
                seqfilter=sfTD,
                **sdopts)
            if len(eTD) < len(newTD):
                rejects += 1
                continue

            eDT = sd2.easyends(
                'DT',
                5,
                number=len(newDT),
                #oldends=oldDTseqs,
                energetics=energetics,
                interaction=interaction,
                chooser=chooser,
                preavail=availendsDT,
                seqfilter=sfDT,
                **sdopts)
            if len(eDT) < len(newDT):
                rejects += 1
                continue

            # FIXME: old ends
            score = sd2.deviation_score(
                [eTD, eDT], all_energetics, devmethod=devmethod)
            if score < min_score:
                min_score = score
                minTD = eTD
                minDT = eDT
                SELOGGER.info("Found set: score {} trial {} rejects {}".format(
                    score, i, rejects))

        newTDseqs = minTD.strs
        newDTseqs = minDT.strs
        info['score'] = float(min_score)

        # Shuffle the lists of end sequences, to ensure that they're
        # random order, and that ends used earlier in the set are not
        # always better than those used later. But only shuffle if
        # there were no templates:
        shuffle(newTDseqs)
        shuffle(newDTseqs)

        # Make sure things are consistent if there are templates:
        for end, s in zip(newDT, newDTseqs):
            ends[end.name].fseq = s
        for end, s in zip(newTD, newTDseqs):
            ends[end.name].fseq = s

        ends.check_consistent()

        # Ensure that the old and new sets have consistent end definitions,
        # and that the tile definitions still fit.
        tileset.ends.merge(ends)
        newtileset.tiles.endlist().merge(ends)

        newendnames = [e.name for e in newTD] + [e.name for e in newDT]
        info['newends'] = newendnames

        # Apply new sequences to tile system.
        newtileset.ends = ends
        if 'info' not in newtileset.keys():
            newtileset['info'] = {}
        if 'end_design' not in newtileset['info'].keys():
            newtileset['info']['end_design'] = []
        if isinstance('end_design', dict):  # convert old
            newtileset['info']['end_design'] = [
                newtileset['info']['end_design']
            ]
        newtileset['info']['end_design'].append(info)

        return newtileset, newendnames

    def create_strand_sequences(tileset,
                                basename='alhambratemp',
                                includes=[
                                    pkg_resources.resource_filename(
                                        __name__, 'peppercomps-j1')
                                ],
                                spurious_pars="verboten_weak=1.5",
                                *options):
        """Given a tileset dictionary with sticky ends sequences, create core sequences
        for tiles, using Pepper.

        Parameters
        ----------

        basename : str, optional
            The base name to use in Pepper (FIXME: why is this needed?).

        includes : list of str paths, optional
            The include paths for Pepper.

        spurious_pars : str
            Command line options for SpuriousDesign

        Returns
        -------

        TileSet
            A TileSet with generated strand sequences for all tiles.
        """

        if not basename:
            import uuid
            basename = str(uuid.uuid4()).replace('-', '')
            # FIXME: is this a valid basename?

        newtileset = copy.deepcopy(tileset)

        newtileset._create_pepper_input_files(basename)

        compiler.compiler(
            basename, [],
            basename + '.pil',
            basename + '.save',
            fixed_file=basename + '.fix',
            includes=includes,
            synth=True)

        spurious_design.design(
            basename,
            infilename=basename + '.pil',
            outfilename=basename + '.mfe',
            verbose=True,
            struct_orient=True,
            tempname=basename + '-temp',
            extra_pars=spurious_pars,
            findmfe=False,
            cleanup=False)

        if 'info' not in newtileset.keys():
            newtileset['info'] = {}
        if 'core' not in newtileset['info'].keys():
            newtileset['info']['core'] = []
        elif isinstance(newtileset['info']['core'], dict):
            newtileset['info']['core'] = [newtileset['info']['core']]

        with open(basename + '-temp.sp') as f:
            a = f.read()
            cdi = {}
            cdi['basename'] = basename
            cdi['score_verboten'] = float(
                re.findall(r'score_verboten\s+score\s+=\s+([+-]?[\d.,]+)', a)[
                    1])
            cdi['score_spurious'] = float(
                re.findall(r'score_spurious\s+score\s+=\s+([+-]?[\d.,]+)', a)[
                    1])
            cdi['score_bonds'] = float(
                re.findall(r'score_bonds\s+score\s+=\s+([+-]?[\d.,]+)', a)[1])
            cdi['score'] = float(
                re.findall(r'weighted score\s+=\s+([+-]?[\d.,]+)', a)[1])
            cdi['spurious_output'] = re.search(
                r"(?<=FINAL\n\n)[\w\W]+weighted.*", a, re.MULTILINE).group(0)

        cdi['time'] = datetime.now(tz=timezone.utc).isoformat()
        newtileset['info']['core'].append(cdi)

        finish.finish(
            basename + '.save',
            designname=basename + '.mfe',
            seqsname=basename + '.seqs',
            strandsname=None,
            run_kin=False,
            cleanup=False,
            trials=0,
            time=0,
            temp=27,
            conc=1,
            spurious=False,
            spurious_time=0)  # FIXME: shouldn't need so many options.

        tileset_with_strands = newtileset._load_pepper_output_files(basename)

        # Ensure:
        tileset.ends.merge(
            tileset_with_strands.tiles.endlist())  # Ends still fit
        for tile in tileset_with_strands.tiles:
            oldtile = tileset.tiles[tile.name]
            if 'fullseqs' in oldtile.keys():
                for old, new in zip(oldtile['fullseqs'], tile['fullseqs']):
                    seq.merge(old, new)  # old tile sequences remain
            assert oldtile.ends == tile.ends

        # Check that old end sequences remain
        tileset['ends'].merge(tileset_with_strands['ends'])

        return tileset_with_strands

    def _create_pepper_input_files(tileset, basename):
        # Are we creating adapters in Pepper?
        if tileset.seed and seeds.seedtypes[tileset['seed']
                                            ['type']].needspepper:
            seedclass = seeds.seedtypes[tileset['seed']['type']]
            createadapts = True
        else:
            createadapts = False

        fixedfile = open(basename + ".fix", 'w')
        # We first need to create a fixed sequence list/file for pepper.
        # Add fixed sticky end and adjacent tile sequences.
        for end in tileset['ends']:
            if 'fseq' not in end.keys():
                continue
            seq = end['fseq'][1:-1]
            if end['type'] == 'TD':
                adj = end['fseq'][-1]
                cadj = end['fseq'][0]  # FIXME: WAS [1], OFF BY ONE!
            elif end['type'] == 'DT':
                adj = end['fseq'][0]
                cadj = end['fseq'][-1]  # FIXME: WAS [1], OFF BY ONE!
            else:
                print("warning! end {} not recognized".format(end['name']))
            fixedfile.write(
                "signal e_{0} = {1}\n".format(end['name'], seq.upper()))
            fixedfile.write(
                "signal a_{0} = {1}\n".format(end['name'], adj.upper()))
            fixedfile.write(
                "signal c_{0} = {1}\n".format(end['name'], cadj.upper()))
            # If we are creating adapter tiles in Pepper, add origami-determined
            # sequences
        if createadapts:
            for i, core in enumerate(seedclass.cores, 1):
                fixedfile.write(
                    "signal origamicore_{0} = {1}\n".format(i, core))

        # Now we'll create the system file in parts.
        importlist = set()
        compstring = ""

        for tile in tileset.tiles:
            e = [[], []]
            for end in tile['ends']:
                if (end == 'hp'):
                    continue
                    # skip hairpins, etc that aren't designed by stickydesign
                e[0].append('e_' + end.replace('/', '*'))
                if end[-1] == '/':
                    a = 'c_' + end[:-1] + '*'
                else:
                    a = 'a_' + end
                e[1].append(a)
            s1 = " + ".join(e[0])
            s2 = " + ".join(e[1])
            tiletype = tile['structure']
            if 'extra' in tile.keys():
                tiletype += '_' + tile['extra']
            compstring += "component {} = {}: {} -> {}\n".format(
                tile['name'], tiletype, s1, s2)
            importlist.add(tiletype)
            if 'fullseqs' in tile.keys():
                fixedfile.write(
                    "structure {}-tile = ".format(tile['name']) + "+".join(
                        [seq.upper() for seq in tile['fullseqs']]) + "\n")

        if createadapts:
            importlist, compstring = seedclass._create_pepper_input_files(
                tileset['seed'], importlist, compstring)

        with open(basename + '.sys', 'w') as sysfile:
            sysfile.write("declare system {}: ->\n\n".format(basename))
            sysfile.write("import " + ", ".join(importlist) + "\n\n")
            sysfile.write(compstring)

    def _load_pepper_output_files(tileset, basename):
        import re

        # Are we creating adapters in Pepper?
        # if seeds.seedtypes[tileset['seed']['type']].needspepper:
        #     seedclass = seeds.seedtypes[tileset['seed']['type']]
        #     createadapts = True

        tset = copy.deepcopy(tileset)

        seqsstring = open(basename + '.seqs').read()

        for tile in tset.tiles:
            pepperstrands = re.compile('strand ' + tile['name'] +
                                       '-([^ ]+) = ([^\n]+)').findall(
                                           seqsstring)
            tile['fullseqs'] = tilestructures.order_pepper_strands(
                pepperstrands)

        #for adapter in tset['seed']['adapters']:
        #    pepperstrands = re.compile('strand ' + adapter['name'] +
        #                               '-([^ ]+) = ([^\n]+)').findall(
        #                                   seqsstring)
        #    adapter['fullseqs'] = tilestructures.order_pepper_strands(
        #        pepperstrands)

        return tset

    def create_guard_strand_sequences(tileset):
        """Given a tileset dictionary with core tile sequences,
        create guard strand sequences.

        Returns
        -------

        TileSet
            A TileSet with generated guard strand sequences.
        """
        tset = tileset.copy()

        for guard in tset['guards']:
            tile = tset.tiles[guard[0]]
            guard.append(wc(tile['fullseqs'][guard[1] - 1]))

        return tset

    def create_adapter_sequence_diagrams(tileset, filename, *options):
        from lxml import etree
        import pkg_resources
        import os.path

        base = etree.parse(
            pkg_resources.resource_stream(
                __name__, os.path.join('seqdiagrambases', 'blank.svg')))
        baseroot = base.getroot()
        pos = 200
        for adapterdef in tileset['seed']['adapters']:

            seedclass = seeds.seedtypes[tileset['seed']['type']]
            group = seedclass.create_adapter_sequence_diagram(adapterdef)

            group.attrib['transform'] = 'translate(0,{})'.format(pos)
            pos += 200
            baseroot.append(group)

        base.write(filename)

    def run_xgrow(self,
                  xgrowparams={},
                  perfect=False,
                  rotate=False,
                  energetics=None,
                  ui=False,
                  output=None,
                  labelsonly=False,
                  onlyreal=True):
        """Run Xgrow for the system.
        
        Parameters
        ----------

        xgrowparams : dict, optional
            Extra Xgrow parameters.  For example, to change Gse and use 2px block, use 
        {'Gse': 9.2, 'block': 2}.

        perfect : bool
            If True, each end binds to its complement by strength 1.  If False,
            stickydesign energetics are used.  Note that TileSets with odd ends
            (eg, with different "ends" that are actually just the same end with
            different adjacent bases, as in the original COUNT system) will likely
            not behave as expected with perfect=True.

        rotate : bool, optional
            If False (Default), tiles are simply converted to Xgrow tiles.  If True,
            rotated/flipped copies of each tile are created in order to simulate valid
            tile orientations other than those intended.

        energetics : stickydesign.Energetics, optional
            The energetics model to use.  If not specified, DEFAULT_ENERGETICS is used.


        output : str or list, optional
            either a string or list, specifying output options.  If a
            string, one of 'final', 'array', or 'trace', corresponding to 'datafile',
            'arrayfile' and 'tracefile', respectively.  If a list of multiple, then
            do those.  These will manage the output, and return the data in usable
            form.

        Returns
        -------
        list or various
            data as requested from output.  See xgrow.run for more information.

        """
        import xgrow
        return xgrow.run(
            self.generate_xgrow_dict(
                perfect=perfect, rotate=rotate, energetics=energetics, labelsonly=labelsonly, onlyreal=onlyreal),
            extraparams=xgrowparams,
            outputopts=output,
            ui=ui)

    def sensitivity_classes(ts, count=False, _maxorder=2):
        return sensitivity.sensitivity_classes(
            ts, count=False, _maxorder=_maxorder)

    def generate_xgrow_dict(ts, perfect=False, rotate=False, energetics=None,
                            labelsonly=False, onlyreal=True):
        """Generate a Xgrow tileset dict.

        Parameters
        ----------

        perfect : bool
            If True, each end binds to its complement by strength 1.  If False,
            stickydesign energetics are used.  Note that TileSets with odd ends
            (eg, with different "ends" that are actually just the same end with
            different adjacent bases, as in the original COUNT system) will likely
            not behave as expected with perfect=True.

        rotate : bool, optional
            If False (Default), tiles are simply converted to Xgrow tiles.  If True,
            rotated/flipped copies of each tile are created in order to simulate valid
            tile orientations other than those intended.

        energetics : stickydesign.Energetics, optional
            The energetics model to use.  If not specified, DEFAULT_ENERGETICS is used.

        Returns
        -------

        dict
            An xgrow-compatible tileset definition, suitable for, eg, xgrow.run.
        """

        # Combine ends and tile-specified adjacents
        newtiles = []
        newends = []
        doubleends = []
        doubles = []
        vdoubleends = []
        vdoubles = []
        ts = copy.deepcopy(ts)
        if ts.seed:
            seedtype = seeds.seedtypes[ts.seed['type']]
            newtiles.append({
                'name':
                'origami',
                'edges': ['origami', 'origami', 'origami', 'origami'],
                'stoic':
                0,
                'color':
                'white'
            })

            atiles = [None] * len(seedtype.cores)
            to_use = []
            # If we have use_adapters, use that, otherwise use every adapter:
            if 'use_adapters' in ts['seed']:
                for tilename in ts['seed']['use_adapters']:
                    try:
                        tile = [
                            x for x in ts['seed']['adapters']
                            if x.get('name') == tilename
                        ][0]
                        to_use.append(tile)
                    except IndexError as e:
                        raise Exception(
                            "Can't find {}".format(tilename)) from e
            else:
                to_use = ts['seed']['adapters']

            for tile in to_use:
                newtile = {}
                if 'ends' in tile.keys():
                    newtile['edges'] = ['origami'] + [
                        re.sub('/', '_c', x) for x in tile['ends']
                    ] + ['origami']
                else:
                    mtile = ts.tiles[tile['tilebase']]
                    newtile['edges'] = ['origami'] + [
                        re.sub('/', '_c', x)
                        for x in
                        mtile.ends[seedtype._mimicadapt[mtile.structure.name]
                                   ['ends']]
                    ] + ['origami']
                newtile['name'] = tile.get('name', '')
                newtile['stoic'] = 0
                newtile['color'] = 'white'
                atiles[tile['loc'] - 1] = newtile
            for tile in atiles:
                if tile:
                    newtiles.append(tile)
                else:
                    newtiles.append({
                        'name': 'emptyadapt',
                        'edges': ['origami', 0, 0, 'origami'],
                        'stoic': 0,
                        'color': 'white'
                    })

        if onlyreal:
            ts.tiles = TileList([x for x in ts.tiles if 'fake' not in x])
                    
        if rotate:
            rotatedtiles = []
            for tile in ts.tiles:
                # only include rotated tiles that aren't identical (handles symmetric tiles)
                trot = [t for t in tile.rotations if (t['ends'] != tile['ends'])]
                td = []
                for i, tr in enumerate(trot):
                    for j in range(0, i):
                        if tr['ends'] == trot[j]['ends']:
                            td.append(i)
                            break
                for i in reversed(td):
                    del(trot[i])
                rotatedtiles += trot
            ts.tiles += rotatedtiles

        for tile in ts.tiles:
            if 'rotation' in tile.keys():
                tile['name'] = tile['name']+'_rot{}'.format(tile['rotation'])
            if (re.match('tile_daoe_3up', tile.structure.name)
                    or re.match('tile_daoe_5up', tile.structure.name)):
                newtile = {}
                newtile['edges'] = [re.sub('/', '_c', x) for x in tile['ends']]
                if 'name' in tile:
                    newtile['name'] = tile['name']
                if 'conc' in tile:
                    newtile['stoic'] = tile['conc']
                if ('color' in tile) and (not labelsonly):
                    newtile['color'] = tile['color']
                elif labelsonly:
                    if 'label' in tile.keys():
                        newtile['color'] = 'white'
                    else:
                        newtile['color'] = 'gray50'
                newtiles.append(newtile)

            if re.match('tile_daoe_doublehoriz', tile.structure.name):
                newtile1 = {}
                newtile2 = {}
                newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:1] ] \
                    + [ tile['name']+'_db' ] \
                    + [ re.sub('/','_c',x) for x in tile['ends'][4:] ]
                newtile2['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][1:4] ] \
                    + [ tile['name']+'_db' ]
                newtile1['name'] = tile['name'] + '_left'
                newtile2['name'] = tile['name'] + '_right'

                doubleends.append(tile['name'] + '_db')
                doubles.append((newtile1['name'], newtile2['name']))

                if 'conc' in tile:
                    newtile1['stoic'] = tile['conc']
                    newtile2['stoic'] = tile['conc']

                if ('color' in tile) and (not labelsonly):                    
                    newtile1['color'] = tile['color']
                    newtile2['color'] = tile['color']
                elif labelsonly:
                    if 'label' in tile.keys():
                        newtile1['color'] = 'white'
                        newtile2['color'] = 'white'
                    else:
                        newtile1['color'] = 'gray50'
                        newtile2['color'] = 'gray50'                        

                newtiles.append(newtile1)
                newtiles.append(newtile2)
            if re.match('tile_daoe_doublevert', tile.structure.name):
                newtile1 = {}
                newtile2 = {}
                newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:2] ] \
                    + [ tile['name']+'_db' ] \
                    + [ re.sub('/','_c',x) for x in tile['ends'][5:] ]
                newtile2['edges'] = [tile['name'] + '_db'] + [
                    re.sub('/', '_c', x) for x in tile['ends'][2:5]
                ]
                newtile1['name'] = tile['name'] + '_top'
                newtile2['name'] = tile['name'] + '_bottom'

                vdoubleends.append(tile['name'] + '_db')
                vdoubles.append((newtile1['name'], newtile2['name']))

                if 'conc' in tile:
                    newtile1['stoic'] = tile['conc']
                    newtile2['stoic'] = tile['conc']

                if ('color' in tile) and (not labelsonly):                    
                    newtile1['color'] = tile['color']
                    newtile2['color'] = tile['color']
                elif labelsonly:
                    if 'label' in tile.keys():
                        newtile1['color'] = 'white'
                        newtile2['color'] = 'white'
                    else:
                        newtile1['color'] = 'gray50'
                        newtile2['color'] = 'gray50'
                
                newtiles.append(newtile1)
                newtiles.append(newtile2)

        newends.append({'name': 'origami', 'strength': 100})

        for end in doubleends:
            newends.append({'name': end, 'strength': 10})
        for end in vdoubleends:
            newends.append({'name': end, 'strength': 10})

        # check for whether we need to use perfect:
        if not perfect:
            for end in ts.allends:
                if (end['type'] in {'TD', 'DT'}) and 'fseq' not in end.keys():
                    perfect = True
                    SELOGGER.warn(
                        "setting perfect=True, because {} has no sequence.".
                        format(end['name']))
                    break

        gluelist = []
        if not perfect:
            glueends = {'DT': [], 'TD': []}
            for end in ts.allends:
                newends.append({'name': end['name'], 'strength': 0})
                newends.append({'name': end['name'] + '_c', 'strength': 0})
                if (end['type'] == 'TD') or (end['type'] == 'DT'):
                    glueends[end['type']].append((end['name'], end['fseq']))

            if energetics:
                ef = energetics
            else:
                ef = DEFAULT_ENERGETICS

            eavg = {}
            for t in ['DT', 'TD']:
                names, fseqs = zip(*glueends[t])
                ea = sd.endarray(fseqs, t)
                eavg[t] = np.average(ef.matching_uniform(ea))
            eavg_combined = (eavg['DT'] + eavg['TD']) / 2.0

            for t in ['DT', 'TD']:
                names, fseqs = zip(*glueends[t])
                allnames = names + tuple(x + '_c' for x in names)
                ea = sd.endarray(fseqs, t)
                ar = sd.energy_array_uniform(ea, ef) / eavg_combined
                for i1, n1 in enumerate(names):
                    for i2, n2 in enumerate(allnames):
                        gluelist.append([n1, n2, max(float(ar[i1, i2]), 0.0)])

        else:
            if 'ends' not in ts.keys():
                ts['ends'] = []
            endsinlist = set(e['name'] for e in ts['ends'])
            endsintiles = set()
            for tile in ts.tiles:
                endsintiles.update(
                    re.sub('/', '', e) for e in tile['ends'] if e != 'hp')
            for end in ts['ends'] + list({'name': e} for e in endsintiles):
                newends.append({'name': end['name'], 'strength': 0})
                newends.append({'name': end['name'] + '_c', 'strength': 0})
                gluelist.append([end['name'], end['name'] + '_c', 1.0])

        newends.append({'name': 'hp', 'strength': 0})

        xga = {}
        xga['doubletiles'] = [list(x) for x in doubles]
        xga['vdoubletiles'] = [list(x) for x in vdoubles]

        xga.update(ts.get('xgrow_options', dict()))

        #if not perfect:
        #    xga['gse_calc_avg'] = eavg_combined

        sts = {
            'tiles': newtiles,
            'bonds': newends,
            'xgrowargs': xga,
            'glues': gluelist
        }

        return sts

    def plot_se_hists(tileset,
                      all_energetics=None,
                      energetics_names=None,
                      title=None,
                      **kwargs):
        """Plot histograms of sticky end energies, using stickydesign.plots.hist_multi.

        Parameters
        ----------

        all_energetics : list of Energetics
            A list of energetics to use.  Defaults to DEFAULT_MULTIMODEL_ENERGETICS.

        energetics_names : list of str
            Names for energetics in all_energetics.  Defaults to DEFAULT_MM_ENERGETICS_NAMES.

        title : str
            Title for the plot.

        **kwargs
            kwargs passed to stickydesign.plots.hist_multi.

        """
        if all_energetics is None:
            all_energetics = DEFAULT_MULTIMODEL_ENERGETICS

        if energetics_names is None:
            energetics_names = DEFAULT_MM_ENERGETICS_NAMES

        if 'ends' in tileset.keys():
            ends = tileset['ends']
        else:
            ends = tileset

        if title is None:
            # FIXME
            title = 'Title'

        td = sd.endarray([x['fseq'] for x in ends if x['type'] == 'TD'], 'TD')

        dt = sd.endarray([x['fseq'] for x in ends if x['type'] == 'DT'], 'DT')
        import stickydesign.plots as sdplots
        return sdplots.hist_multi([td, dt], all_energetics, energetics_names,
                                  title, **kwargs)

    def plot_se_lv(self,
                   all_energetics=None,
                   energetics_names=None,
                   pltcmd=None,
                   title=None,
                   **kwargs):

        if all_energetics is None:
            all_energetics = DEFAULT_MULTIMODEL_ENERGETICS

        if energetics_names is None:
            energetics_names = DEFAULT_MM_ENERGETICS_NAMES
        import stickydesign.plots as sdplots
        m, s = sdplots._multi_data_pandas(self.ends.to_endarrays(),
                                          all_energetics, energetics_names)

        import seaborn as sns
        import matplotlib.pyplot as plt

        if pltcmd is None:
            pltcmd = sns.lvplot

        pltcmd(data=m, **kwargs)
        pltcmd(data=s, marker='x', **kwargs)
        if title:
            plt.title(title)
        plt.ylabel("Energy (kcal/mol)")

    def plot_adjacent_regions(tileset, energetics=None):
        """
        Plots the strength of double-stranded regions in DX tiles adjacent 
        to sticky ends.

        Parameters
        ----------

        energetics : stickydesign.Energetics
            The energetics to use.  Defaults to DEFAULT_REGION_ENERGETICS.
        """

        if energetics is None:
            energetics = DEFAULT_REGION_ENERGETICS

        regions = [t.structure._side_bound_regions(t) for t in tileset.tiles]
        regions = [[x.lower() for x in y] for y in regions]
        allregions = sum(regions, [])
        count = [[Counter(x) for x in y] for y in regions]
        gc_count = [[x['g'] + x['c'] for x in c] for c in count]
        gc_counts = sum(gc_count, [])

        ens = energetics.matching_uniform(sd.endarray(allregions, 'DT'))
        from matplotlib import pylab
        pylab.figure(figsize=(10, 4))
        pylab.subplot(121)
        pylab.hist(
            gc_counts,
            bins=np.arange(min(gc_counts) - 0.5, max(gc_counts) + 0.5))
        pylab.title('G/C pairs in arms')
        pylab.ylabel('# of 8 nt arms')
        pylab.xlabel('# of G/C pairs')
        pylab.subplot(122)
        pylab.hist(ens)
        pylab.title('ΔG, T=33, no coaxparams/danglecorr')
        pylab.ylabel('# of 8 nt regions')
        pylab.xlabel('stickydesign ΔG')
        pylab.suptitle('8 nt end-adjacent region strengths')

    def plot_side_strands(tileset, energetics=None):
        """
        Plots the binding strength of short strands in DX tiles.

        Parameters
        ----------

        energetics : stickydesign.Energetics
            The energetics to use.  Defaults to DEFAULT_REGION_ENERGETICS.
        """

        if energetics is None:
            energetics = DEFAULT_REGION_ENERGETICS

        regions = [t.structure._short_bound_full(t) for t in tileset.tiles]
        regions = [[x.lower() for x in y] for y in regions]
        allregions = sum(regions, [])
        count = [[Counter(x) for x in y] for y in regions]
        gc_count = [[x['g'] + x['c'] for x in c] for c in count]
        gc_counts = sum(gc_count, [])

        ens = energetics.matching_uniform(sd.endarray(allregions, 'DT'))
        from matplotlib import pylab
        pylab.figure(figsize=(10, 4))
        pylab.subplot(121)
        pylab.hist(
            gc_counts,
            bins=np.arange(min(gc_counts) - 0.5, max(gc_counts) + 0.5))
        pylab.title('G/C pairs in arms')
        pylab.ylabel('# of 8 nt arms')
        pylab.xlabel('# of G/C pairs')
        pylab.subplot(122)
        pylab.hist(ens)
        pylab.title('ΔG, T=33, no coaxparams/danglecorr')
        pylab.ylabel('# of 16 nt regions')
        pylab.xlabel('stickydesign ΔG')
        pylab.suptitle('16 nt arm region strengths')

    def reduce_ends(tileset, preserve, tries, threads):
        pass

    def reduce_tiles(tileset, preserve=('s22','ld'), tries=10, threads=1, returntype='equiv', best=1, key=None, initequiv=None):
        """
        Apply tile reduction algorithm, preserving some set of properties, and using a multiprocessing pool.

        Parameters
        ----------
        tileset: TileSet  
            The system to reduce. 

        preserve: a tuple or list of strings, optional
            The properties to preserve.  Currently supported are 's1' for first order
            sensitivity, 's2' for second order sensitivity, 's22' for two-by-two sensitivity,
            'ld' for small lattice defects, and 'gs' for glue sense (to avoid spurious
            hierarchical attachment).  Default is currently ('s22', 'ld').

        tries: int, optional
            The number of times to run the algorithm.

        threads: int, optional
            The number of threads to use (using multiprocessing).

        returntype: 'TileSet' or 'equiv' (default 'equiv')
            The type of object to return.  If 'equiv', returns an array of glue equivalences
            (or list, if best != 1) that can be applied to the tileset with apply_equiv, or used 
            for further reduction.  If 'TileSet', return a TileSet with the equiv already applied
            (or a list, if best != 1).

        best: int or None, optional
            The number of systems to return.  If 1, the result will be returned
            directly; if k > 1, a list will be returned of the best k results (per cmp);
            if k = None, a list of *all* results will be returned, sorted by cmp. (default 1)

        key: function (ts, equiv1, equiv2) -> some number/comparable
            A comparison function for equivs, to sort the results. FIXME: documentation needed.
            Default (if None) here is to sort by number of glues in the system, regardless of number 
            of tiles.

        initequiv: equiv
            If provided, the equivalence array to start from.  If None, start from the tileset without
            any merged glues.

        Returns
        -------
        reduced: single TileSet or equiv, or list
            The reduced system/systems
        """
        return fastreduce.reduce_tiles(tileset, preserve, tries, threads, returntype, best, key, initequiv)

    def reduce_ends(tileset, preserve=('s22','ld'), tries=10, threads=1, returntype='equiv', best=1, key=None, initequiv=None):
        """
        Apply end reduction algorithm, preserving some set of properties, and using a multiprocessing pool.

        Parameters
        ----------
        tileset: TileSet  
            The system to reduce. 

        preserve: a tuple or list of strings, optional
            The properties to preserve.  Currently supported are 's1' for first order
            sensitivity, 's2' for second order sensitivity, 's22' for two-by-two sensitivity,
            'ld' for small lattice defects, and 'gs' for glue sense (to avoid spurious
            hierarchical attachment).  Default is currently ('s22', 'ld').

        tries: int, optional
            The number of times to run the algorithm.

        threads: int, optional
            The number of threads to use (using multiprocessing).

        returntype: 'TileSet' or 'equiv' (default 'equiv')
            The type of object to return.  If 'equiv', returns an array of glue equivalences
            (or list, if best != 1) that can be applied to the tileset with apply_equiv, or used 
            for further reduction.  If 'TileSet', return a TileSet with the equiv already applied
            (or a list, if best != 1).

        best: int or None, optional
            The number of systems to return.  If 1, the result will be returned
            directly; if k > 1, a list will be returned of the best k results (per cmp);
            if k = None, a list of *all* results will be returned, sorted by cmp. (default 1)

        key: function (ts, equiv1, equiv2) -> some number/comparable
            A comparison function for equivs, to sort the results. FIXME: documentation needed.
            Default (if None) here is to sort by number of glues in the system, regardless of number 
            of tiles.

        initequiv: equiv
            If provided, the equivalence array to start from.  If None, start from the tileset without
            any merged glues.

        Returns
        -------
        reduced: single TileSet or equiv, or list
            The reduced system/systems
        """
        return fastreduce.reduce_ends(tileset, preserve, tries, threads, returntype, best, key, initequiv)

    
    def latticedefects(ts, direction='e', depth=2, pp=True, rotate=False):
        from . import latticedefect
        return latticedefect.latticedefects(
            ts, direction=direction, depth=depth, pp=pp, rotate=rotate)

    def apply_equiv(ts, equiv):
        return fastreduce._FastTileSet(ts).applyequiv(ts, equiv)

RoundTripRepresenter.add_representer(TileSet,
                                     RoundTripRepresenter.represent_dict)



# New sensitivity code.
from .tilestructures import tile_daoe_single
from .tiles import TileList, Tile
from collections import Counter
from .util import comp, GlueMergeSpec, TileMergeSpec


def _fakesingle(tile):
    if not tile.structure.double:
        return TileList([tile])
    ret = TileList()
    c = ['', '/']
    cn = ['_db_a', '_db_b']
    for qq, es in enumerate(tile.structure.singleends):
        ft = Tile()
        ft.structure = tile_daoe_single()
        ft.structure._endlocs = [None, None, None, None]
        ft['ends'] = []
        ft['name'] = tile.name + cn[qq]
        if 'input' in tile.keys():
            ft['input'] = []
        ft.structure._endtypes = []
        for i in es:
            if i is not None:
                ft['ends'].append(tile['ends'][i])
                ft.structure._endtypes.append(tile.structure._endtypes[i])
                if 'input' in tile.keys():
                    ft['input'].append(tile['input'][i])
            else:
                ft['ends'].append(tile.name + '_db' + c[qq])
                ft.structure._endtypes.append('fakedouble')
                if 'input' in tile.keys():
                    ft['input'].append(0)
        ret.append(ft)
    return ret


def _fakesingles(tiles):
    return sum((_fakesingle(x) for x in tiles), TileList())


_rev = [2, 3, 0, 1]


def sensitivity_profiles_fakesingles(tileset,
                                     _maxorder=2,
                                     ms=GlueMergeSpec([]),
                                     tms=TileMergeSpec([]),
                                     oldclasses=None,
                                     checks=[],
                                     stopfirst=False):
    singles = _fakesingles(tileset.tiles)
    rotatedsingles = singles + _fakesingles(
        sum([x.rotations for x in tileset.tiles], TileList()))

    spairs = {'1GO': set(), '1NGO': set(), '2NGO': set(), '2GO': set()}
    if _maxorder >= 3:
        spairs['22GO'] = set()
        spairs['22NGO'] = set()

    occ = dict()

    # Convert tile names
    if oldclasses:
        for k, v in oldclasses.items():
            occ[k] = {(tms._cm.get(t1, t1), tms._cm.get(t2, t2))
                      for t1, t2 in v}

    for t1 in singles:
        # Iterate glues searching for a shared glue
        for i in range(0, 4):
            # If the glue is fake, then continue
            if t1.structure._endtypes[i] in {'fakedouble', 'hairpin'}:
                continue
            for t2 in rotatedsingles:
                if not ms.eq(t1['ends'][i], t2['ends'][i]):
                    # not shared glue
                    continue
                if all(ms.eq(x, y) for x, y in zip(t1['ends'], t2['ends'])):
                    # NoIO(t1) = NoIO(t2)
                    continue
                # shared glue and not same tile: search for
                # viable unshared glue
                for j in range(0, 4):
                    if i == j:
                        # same edge
                        continue
                    if ms.eq(t1['ends'][j], t2['ends'][j]):
                        continue
                    if t1.structure._endtypes[j] != t2.structure._endtypes[j]:
                        continue
                    if t1.structure._endtypes[j] in {'fakedouble', 'hairpin'}:
                        continue
                    spairs['1NGO'].add((t1.name, t2.name))
                    go1 = False
                    if t1['input'][i] and t1['input'][j]:
                        go1 = True
                        if (stopfirst and '1GO' in checks and
                            ((tms._cm.get(t1.name, t1.name),
                              tms._cm.get(t2.name, t2.name))
                             not in occ['1GO'])):
                            return (t1, t2)
                        spairs['1GO'].add((t1.name, t2.name))
                    for k in (set(range(0, 4)) - {i, j}):
                        if (t1.structure._endtypes[k] == 'hairpin') or (
                                t2.structure._endtypes[k] == 'hairpin'):
                            continue
                        ec1 = comp(t1['ends'][k])
                        ec2 = comp(t2['ends'][k])
                        kc = _rev[k]
                        for t12 in singles:
                            if not ms.eq(t12['ends'][kc], ec1):
                                continue
                            for t22 in rotatedsingles:
                                if not ms.eq(t22['ends'][kc], ec2):
                                    continue
                                for m in range(0, 4):
                                    if m == kc:
                                        continue
                                    if t12.structure._endtypes[m] in {
                                            'fakedouble', 'hairpin'
                                    }:
                                        continue
                                    if not ms.eq(t22['ends'][m],
                                                 t12['ends'][m]):
                                        continue
                                    spairs['2NGO'].add((t1.name, t2.name))
                                    if t12['input'][m] and t12['input'][kc] and go1:
                                        spairs['2GO'].add((t1.name, t2.name))
                                        if (stopfirst and '2GO' in checks and
                                            ((tms._cm.get(t1.name, t1.name),
                                              tms._cm.get(t2.name, t2.name))
                                             not in occ['2GO'])):
                                            return (t1, t2)
                                    if _maxorder > 2:
                                        for k2 in (set(range(0, 4)) - {
                                                i, j, k
                                        }):  # should just be one number!
                                            # hairpin? then continue.
                                            if (t1.structure._endtypes[k2] ==
                                                    'hairpin'
                                                ) or (
                                                    t2.structure._endtypes[k2]
                                                    == 'hairpin'):
                                                continue
                                            # fakedouble on either: then assume we're doomed: (FIXME)
                                            if (t1.structure._endtypes[k2] ==
                                                    'fakedouble'
                                                ) or (
                                                    t2.structure._endtypes[k2]
                                                    == 'fakedouble'):
                                                spairs['22NGO'].add((t1.name,
                                                                     t2.name))
                                                spairs['22GO'].add((t1.name,
                                                                    t2.name))
                                                if (stopfirst and '22GO' in checks
                                                    and ((tms._cm.get(t1.name, t1.name),
                                                          tms._cm.get(t2.name, t2.name))
                                                         not in occ['22GO'])):
                                                    return (t1, t2)
                                                continue

                                            ec1_3 = comp(t1['ends'][k2])
                                            ec2_3 = comp(t2['ends'][k2])
                                            kc2 = _rev[k2]
                                            for t13 in singles:
                                                if not ms.eq(
                                                        t13['ends'][kc2],
                                                        ec1_3):
                                                    continue
                                                for t23 in rotatedsingles:
                                                    if not ms.eq(
                                                            t23['ends'][kc2],
                                                            ec2_3):
                                                        continue
                                                    for m2 in range(0, 4):
                                                        if m2 == kc2:
                                                            continue
                                                        if not ms.eq(
                                                                t23['ends']
                                                            [m2], t13['ends']
                                                            [m2]):
                                                            continue
                                                        if t13.structure._endtypes[
                                                                m2] in {
                                                                    'fakedouble',
                                                                    'hairpin',
                                                                }:
                                                            continue
                                                        spairs['22NGO'].add(
                                                            (t1.name, t2.name))
                                                        if (t12['input'][m] and
                                                            t12['input'][kc] and
                                                            t13['input'][m2] and
                                                            t13['input'][kc2] and
                                                            go1):
                                                            if (stopfirst and '22GO' in checks
                                                                and ((tms._cm.get(t1.name, t1.name),
                                                                      tms._cm.get(t2.name, t2.name))
                                                                     not in occ['22GO'])):
                                                                return (t1, t2)
                                                            spairs['22GO'].add(
                                                                (t1.name,
                                                                 t2.name))

    if not stopfirst:
        return spairs
    else:
        return False



from .tilestructures import check_edotparen_sequence, check_edotparen_consistency
import copy


class seed_base:
    def create_adapter_sequence_diagram(self, adapterdef):
        # At the moment, there are only three potential structures, so I'm just
        # going to put them here.
        ttype = 'tile_adapter_5up'
        if 'extra' in adapterdef.keys():
            ttype += '_' + adapterdef['extra']

        from lxml import etree
        import pkg_resources
        import os
        base_svg = etree.parse(
            pkg_resources.resource_stream(
                __name__, os.path.join('seqdiagrambases', ttype + '.svg')))

        try:
            s = adapterdef['seqs'][1]
            l = adapterdef['seqs'][0]
        except KeyError:
            s = adapterdef['fullseqs'][1]
            l = adapterdef['fullseqs'][0]
        name = adapterdef['name']
        if 'extra' not in adapterdef.keys():
            strings = [(s[0:5] + '--' + s[5:13])[::-1],
                       s[13:21] + '--' + s[21:26],
                       (l[0:8] + '--' + l[8:16])[::-1], l[16:24][::-1],
                       l[24:32], l[32:40] + '--' + l[40:48],
                       adapterdef['ends'][0], adapterdef['ends'][1], name]
        elif adapterdef['extra'] == '1h':
            hp = s[0:13]
            s = s[13:]
            strings = [
                hp[0:5] + '-' + hp[5:7] + '-' + hp[7:9],
                (hp[9:11] + '-' + hp[11:13] + '-' + s[0:5] + '--' +
                 s[5:13])[::-1], s[13:21] + '--' + s[21:26],
                (l[0:8] + '--' + l[8:16])[::-1], l[16:24][::-1], l[24:32],
                l[32:40] + '--' + l[40:48], adapterdef['ends'][1], name
            ]
        elif adapterdef['extra'] == '2h':
            strings = [
                (s[0:5] + '--' + s[5:13])[::-1],
                s[13:21] + '--' + s[21:26] + '-' + s[26:28] + '-' + s[28:30],
                (s[30:32] + '-' + s[32:34] + '-' + s[34:39])[::-1],
                (l[0:8] + '--' + l[8:16])[::-1], l[16:24][::-1], l[24:32],
                l[32:40] + '--' + l[40:48], adapterdef['ends'][0], name
            ]

        texts = base_svg.findall("//{http://www.w3.org/2000/svg}text")

        for text, string in zip(texts, strings):
            text.text = string

        return base_svg.xpath(
            '/svg:svg/svg:g', namespaces={'svg':
                                          'http://www.w3.org/2000/svg'})[0]


class seed_tileadapts(seed_base):
    needspepper = False

    _mimicadapt = {
        'tile_daoe_3up': {
            'strand': 3,
            'ends': slice(1, 3),
            'structure': 'tile_adapter_3up'
        },
        'tile_daoe_5up': {
            'strand': 3,
            'ends': slice(1, 3),
            'structure': 'tile_adapter_5up'
        },
        'tile_daoe_5up_2h': {
            'strand': 3,
            'ends': slice(1, 3),
            'structure': 'tile_adapter_5up_1h'
        },
        'tile_daoe_5up_3b': {
            'strand': 3,
            'ends': slice(1, 3),
            'structure': 'tile_adapter_5up_2h'
        },
        'tile_daoe_doublehoriz_35up_2h3h': {
            'strand': 5,
            'ends': slice(2, 4),
            'structure': 'tile_adapter_5up_1h'
        },
        'tile_daoe_doublehoriz_35up_1h2i': {
            'strand': 5,
            'ends': slice(2, 4),
            'structure': 'tile_adapter_5up'
        },
        'tile_daoe_doublehoriz_35up': {
            'strand': 5,
            'ends': slice(2, 4),
            'structure': 'tile_adapter_5up'
        },
        'tile_daoe_doublevert_35up_4h5h': {
            'strand': 5,
            'ends': slice(2, 4),
            'structure': 'tile_adapter_5up_2h'
        }
    }

    def create_adapter_sequences(self, tileset):
        tset = copy.deepcopy(tileset)

        for adapter in tset['seed']['adapters']:
            # It's not a tile, it's a...
            mimic = tileset.tiles[adapter['tilebase']]

            structinfo = self._mimicadapt[mimic['structure']]

            if (('ends' in adapter.keys())
                    and (adapter['ends'] != mimic.ends[structinfo['ends']])):
                raise ValueError(
                    "adapter {} and tile base {} ends don't match: adapter has {}, tile has {}".
                    format(adapter['name'], mimic.name, adapter['ends'],
                           mimic.ends[structinfo['ends']]))
            adapter_strand_short = mimic.strands[-1]
            tile_long_strand = mimic.strands[-2]

            adapter['seqs'] = [  # FIXME: better to use short strand here
                tile_long_strand[0:8] + self.cores[adapter['loc'] - 1] +
                tile_long_strand[40:48], adapter_strand_short
            ]

            adapter['structure'] = structinfo['structure']
        return tset

    def check_consistent(self, tileset):
        for adapt in tileset['seed']['adapters']:
            mimic = tileset.tiles[adapt['tilebase']]
            structinfo = self._mimicadapt[mimic['structure']]
            if adapt.get('structure'):
                assert adapt.get('structure') == structinfo['structure']
            else:
                adapt['structure'] = structinfo['structure']  # FIXME: should be in init code
            if 'ends' in adapt.keys():
                assert adapt['ends'] == mimic.ends[structinfo['ends']]


    edotparens = {
        'tile_adapter_5up': "8(32.8(+5.16)5.",
        'tile_adapter_5up_2h': "8(32.8(+5.16)7(4.7)",
        'tile_adapter_5up_1h': "8(32.8(+7(4.7)16)5."
    }

    def check_sequence(self, tileset):
        for p in self.edotparens.values():
            check_edotparen_consistency(p)
        self.check_consistent(tileset)
        for adapt in tileset['seed']['adapters']:
            if 'seqs' not in adapt.keys():
                continue
            seqs = "+".join(adapt['seqs'])
            check_edotparen_sequence(self.edotparens[adapt['structure']], seqs)
            assert adapt['seqs'][0][8:(8 + 32)] == self.cores[adapt['loc'] - 1]


class tallrect_tileadapts(seed_tileadapts):
    cores = [
        'CAGGAACGGTACGCCATTAAAGGGATTTTAGA', 'CTACATTTTGACGCTCACGCTCATGGAAATAC',
        'CCAGCAGAAGATAAAAAATACCGAACGAACCA', 'GCCGTCAATAGATAATCAACTAATAGATTAGA',
        'ACTTCTGAATAATGGATGATTGTTTGGATTAT', 'GAAGATGATGAAACAAAATTACCTGAGCAAAA',
        'CATAGGTCTGAGAGACGTGAATTTATCAAAAT', 'GAAAAAGCCTGTTTAGGGAATCATAATTACTA',
        'ACGCGCCTGTTTATCAGTTCAGCTAATGCAGA', 'GCTTATCCGGTATTCTAAATCAGATATAGAAG',
        'AACGTCAAAAATGAAAAAACGATTTTTTGTTT', 'GCAGATAGCCGAACAATTTTTAAGAAAAGTAA',
        'AGACAAAAGGGCGACAGGTTTACCAGCGCCAA', 'GCGTCAGACTGTAGCGATCAAGTTTGCCTTTA',
        'GTCAGACGATTGGCCTCAGGAGGTTGAGGCAG', 'TGAAAGTATTAAGAGGCTATTATTCTGAAACA'
    ]


class triangle_side2(seed_tileadapts):
    # Note: this list is currently reversed because
    # the cadnano file for the triangle has 5' ends of
    # the last staple above 3' ends, while our adapter
    # scheme has 3' ends above 5' ends.
    cores = list(
        reversed([
            'AGAGAGTACCTTTAATCCAACAGGTCAGGATT',
            'TAAGAGGAAGCCCGAAATTGCATCAAAAAGAT',
            'CCCCCTCAAATGCTTTTAAATATTCATTGAAT',
            'TACCAGACGACGATAATATCATAACCCTCGTT',
            'CGTTAATAAAACGAACTGGGAAGAAAAATCTA',
            'TAACAAAGCTGCTCATATTACCCAAATCAACG',
            'ATTGTGTCGAAATCCGTATCATCGCCTGATAA',
            'CGAGGGTAGCAACGGCAAAGACAGCATCGGAA',
            'TCTCCAAAAAAAAGGCTTTTTCACGTTGAAAA',
            'TTTCGTCACCAGTACAGTACCGTAACACTGAG',
            'TCAGTACCAGGCGGATATTAGCGGGGTTTTGC',
            'GATACAGGAGTGTACTATACATGGCTTTTGAT',
            'ACCAGAGCCGCCGCCACGCCACCAGAACCACC',
            'GCGTTTGCCATCTTTTCATAGCCCCCTTATTA',
            'ATGAAACCATCGATAGGCCGGAAACGTCACCA',
            'ATCACCGTCACCGACTTCATTAAAGGTGAATT'
        ]))


class tallrect_endadapts(seed_base):
    needspepper = True
    cores = tallrect_tileadapts.cores

    def _create_pepper_input_files(self, seeddef, importlist, compstring):
        for adapter in seeddef['adapters']:
            adaptertype = 'tile_adapter_5up'
            e = [[], []]
            e[0].append('origamicore_{0}'.format(adapter['loc']))
            for end in adapter['ends']:
                if (end == 'hp'):
                    continue  # skip hairpins, etc that aren't designed by stickydesign
                e[0].append('e_' + end.replace('/', '*'))
                if end[-1] == '/':
                    a = 'c_' + end[:-1] + '*'
                else:
                    a = 'a_' + end
                e[1].append(a)
            s1 = " + ".join(e[0])
            s2 = " + ".join(e[1])
            if 'extra' in adapter.keys():
                adaptertype += '_' + adapter['extra']
            compstring += "component {} = {}: {} -> {}\n".format(
                adapter['name'], adaptertype, s1, s2)
            importlist.add(adaptertype)
        return (importlist, compstring)

class bigseed(seed_tileadapts):
    cores = ["A"]*37

seedtypes = {
    'tallrect_tileadapts': tallrect_tileadapts(),
    'tallrect_endadapts': tallrect_endadapts(),
    'triangle_side2': triangle_side2(),
    'bigseed': bigseed()
}
seedtypes['longrect'] = seedtypes['tallrect_tileadapts']
seedtypes['longrect_old'] = seedtypes['tallrect_endadapts']



import warnings
from ruamel.yaml.comments import CommentedMap
from .util import NamedList
import copy
from peppercompiler.DNA_classes import wc
from . import seq
import stickydesign as sd
import stickydesign.stickydesign2 as sd2
from .util import DEFAULT_SD2_MULTIMODEL_ENERGETICS, DEFAULT_MM_ENERGETICS_NAMES


class End(CommentedMap):
    def __str__(self):
        if self.fseq:
            if self.etype == 'DT':
                s = self.seq[0]+'-'+self.seq[1:]
                c = self.comp[0]+'-'+self.comp[1:]
            elif self.etype == 'TD':
                s = self.seq[:-1]+'-'+self.seq[-1]
                c = self.comp[:-1]+'-'+self.comp[-1]
            return "<end {} ({}{}): {} | {}>".format(
                self['name'], self['type'], len(self.seq), s, c)
        else:
            return "<end {} ({})>".format(
                self['name'], self.get('type', '?'))

    def fseq():
        """The fseq / full sequence of the End, including the adjacent base on the end
        and the complement of the adjacent base on the complement."""
        def fget(self):
            return self.get('fseq', None)
    
        def fset(self, value):
            if self.get('fseq', None) and len(value) != len(self['fseq']):
                warnings.warn("Changing end length")
            self['fseq'] = value
    
        def fdel(self):
            del self['fseq']
        return locals()
    fseq = property(**fseq())

    def name():

        def fget(self):
            return self['name']
    
        def fset(self, value):
            self['name'] = value
    
        return locals()
    name = property(**name())

    def use():
        def fget(self):
            return self.get('use', 0b0)
    
        def fset(self, value):
            self['use'] = value
    
        return locals()
    use = property(**use())

    def strength():
        def fget(self):
            return self.get('strength', 1)
    
        def fset(self, value):
            self['strength'] = value
    
        return locals()
    strength = property(**strength())
    
    
    @property
    def seq(self):
        """The end sequence (of just the end) of the End, as a string."""
        if not self.fseq:
            return None
        if self.etype == 'TD':
            return self.fseq[1:]
        elif self.etype == 'DT':
            return self.fseq[:-1]

    @property
    def comp(self):
        """The complement end sequences of the End, as a string."""
        if not self.fseq:
            return None
        if self.etype == 'TD':
            return wc(self.fseq[:-1].upper()).lower()
        elif self.etype == 'DT':
            return wc(self.fseq[1:].upper()).lower()

    @property
    def etype(self):
        """The end type of the end."""
        return self['type']

    def merge(end1, end2):
        """Given ends end1 and end2, assuming they describe the same sticky
    end, merge them into a single end, combining information from each and
    enforcing that the two input ends consistently make a single output
    end.
        """
        # Of things in the ends: fseq might need special care, everything
        # else just needs to match.
        out = copy.deepcopy(end1)
        for i, v in end2.items():
            if i in out.keys() and i == 'fseq':
                # we merge sequences
                out[i] = seq.merge(out[i], v)
            elif i in out.keys() and i == 'use':
                out[i] = out[i] | v
            elif i in out.keys() and out[i] != v:
                # we insisted all others must be equal
                raise ValueError(
                    "end1 has {}={}, end2 has {}={}".format(i, out[i], i, v))
            elif i not in out.keys():
                out[i] = copy.deepcopy(v)
        return out

class EndList(NamedList):
    """A list of End instances, which can be merged together, converted to
       stickydesign.endarray instances, etc."""
    def __init__(self, val=[]):
        """Create an EndList instance.

        Parameters
        ----------

        val : iterable
            an enumerable iterable of objects that can initialize End instances.
        """
        NamedList.__init__(self, val)
        for i, end in enumerate(self):
            self[i] = End(end)

    def to_endarrays(self):
        """Return stickydesign endarrays of each type of (non-hairpin) end.

        Returns
        -------

        list of stickydesign.endarray
            the endarrays of ends in the system.
        """
        
        endtypes = {x['type'] for x in self}
        endtypes = endtypes - {'hp', 'hairpin'}
        return list(sd.endarray([x['fseq'] for x in self if x['type'] == y], y)
                     for y in endtypes)

    @property
    def _epas(self):
        """Return stickydesign2 EndPairArrays"""
        endtypes = {x['type'] for x in self}
        endtypes = endtypes - {'hp', 'hairpin'}
        epat = {'TD': sd2.EndPairArrayTD, 'DT': sd2.EndPairArrayDT}
        return list(epat[y]([x.fseq for x in self if x['type'] == y])
                    for y in endtypes)


    @property
    def _epans(self):
        """Return stickydesign2 EndPairArrays"""
        endtypes = {x['type'] for x in self}
        endtypes = endtypes - {'hp', 'hairpin'}
        epat = {'TD': sd2.EndPairArrayTD, 'DT': sd2.EndPairArrayDT}
        return list(list([x.name for x in self if x['type'] == y])
                    for y in endtypes)

    def _pandas_data(self, models=DEFAULT_SD2_MULTIMODEL_ENERGETICS,
                     modelnames=DEFAULT_MM_ENERGETICS_NAMES):
        import stickydesign.stickydesign2.plots as s2pl
        return s2pl._pandas_data(self._epas, self._epans, models, modelnames)
    
    def merge(endlist1, endlist2, fail_immediate=False,
              in_place=False):
        """\
        Given end lists `endlist1` and `endlist2`, merge the two lists, using
        `merge_ends` to merge any named ends that are present in both.

        Parameters
        ----------

        endlist1: NamedList of sticky ends.

        endlist2: NamedList OR just a list of sticky ends.  If it just a list,
            which may have multiple copies of the same named sticky end, each
            sticky end in order is merged.

        fail_immediate: (default False) if True, fail immediately on a
        failed merge, passing through the ValueError from the end merge.
        If False, finish merging the two lists, then raise a
        ValueError listing *all* ends that failed to merge.

        in_place: (default False) if True, do merging in place in endlist1.
        Note the merged and added ends from endlist2 will be copies regardless.

        Returns
        -------
        
        EndList
            a merged list of sticky ends


        Exceptions
        ----------

        ValueError: In the event of a failed merge, when named ends
        cannot be merged.  If fail_immediate is True, then this is passed
        through from merge_ends.  If fail_immediate is False, then the
        ValueError has the following args: ("message", [exceptions],
        [failed_name,failed_end1,failed_end2) ...],out)

        """
        # Check consistency of each NamedList
        endlist1.check_consistent()
        try:
            endlist2.check_consistent()
        except AttributeError:
            pass

        if not in_place:
            out = copy.deepcopy(endlist1)
        else:
            out = endlist1

        exceptions = []
        errors = []

        for end in endlist2:
            if end.name in out.keys():
                try:
                    out[end.name] = out[end.name].merge(end)
                except ValueError as e:
                    if fail_immediate:
                        raise e
                    else:
                        exceptions.append(e)
                        errors.append((end.name, out[end.name], end))
            else:
                out.append(copy.deepcopy(end))

        if errors:
            errorstring = " ".join([e[0] for e in errors])
            raise ValueError("Errors merging {}".format(errorstring),
                             exceptions, errors, out)

        return out


from ruamel.yaml.representer import RoundTripRepresenter
RoundTripRepresenter.add_representer(EndList,
                                     RoundTripRepresenter.represent_list)
RoundTripRepresenter.add_representer(End,
                                     RoundTripRepresenter.represent_dict)



from collections import namedtuple
from copy import copy
import numpy as np
from . import tilestructures
from .tiles import TileList
from .ends import End
from random import shuffle
from . import util
from . import fastlatticedefect as fld
import logging

FTile = namedtuple("FTile", ("color", "use", "glues", "name", "used",
                             "structure", "dfake", "sfake"))

FTilesArray = namedtuple("FTilesArray",
                         ("color", "use", "glues", "name", "used", "structure",
                          "dfake", "sfake"))

TAU = 2


class FGlueList():
    def __init__(self, glues):
        self.name = []
        self.strength = []
        self.structure = []
        self.complement = []
        self.use = []
        self.tonum = {}
        # self.use = []
        for i, g in enumerate(glues):
            self.name.append(g.name)
            self.name.append(g.name + '/')
            self.complement.append(2 * i + 1)
            self.complement.append(2 * i)
            self.structure.append(g.etype)
            self.structure.append(g.etype)
            self.tonum.update({g.name: 2 * i, g.name + '/': 2 * i + 1})
            self.strength.append(g.strength)
            self.strength.append(g.strength)
            self.use.append(g.use)
            self.use.append((g.use // 2) + (g.use % 2) * 2)
        self.name = np.array(self.name)
        self.strength = np.array(self.strength)
        self.structure = np.array(self.structure)
        self.complement = np.array(self.complement)
        self.use = np.array(self.use)

    def blankequiv(self):
        return np.arange(0, len(self.name))

    def iseq(self, equiv, a, b):
        return equiv[a] == equiv[b]

    def domerge(self, equiv, a, b, preserveuse=False):
        if self.structure[a] != self.structure[b]:
            raise ValueError("structure")
        elif self.strength[a] != self.strength[b]:
            raise ValueError("strength")
        elif equiv[a] == equiv[self.complement[b]]:
            raise ValueError("self-comp")
        elif preserveuse and (self.use[a] != self.use[b]):
            raise ValueError("use")
        else:
            equiv = copy(equiv)
            newg, oldg = sorted((equiv[a], equiv[b]))
            newc, oldc = sorted((equiv[self.complement[a]],
                                 equiv[self.complement[b]]))
            equiv[equiv == oldg] = newg
            equiv[equiv == oldc] = newc
        return equiv


U_UNUSED = -1
U_INPUT = 1
U_OUTPUT = 0


class FTileList():
    def __init__(self, tiles, gluelist):
        self.tiles = []
        self.totile = {}
        for t in tiles:
            glues = np.array([gluelist.tonum[x] for x in t.ends])
            if 'fake' in t.keys():
                continue
            if 'input' not in t.keys():
                used = False
                use = np.array([U_UNUSED for _ in t.ends])
            else:
                used = True
                use = np.array([int(x) for x in t['input']])
            color = 'label' in t.keys()
            self.tiles.append(
                FTile(
                    name=t.name,
                    color=color,
                    use=use,
                    glues=glues,
                    used=used,
                    structure=t.structure,
                    dfake=0,
                    sfake=0))
            assert t.name not in self.totile.keys()
            self.totile[t.name] = self.tiles[-1]
        stiles = []
        htiles = []
        vtiles = []

        for t in self.tiles:
            if isinstance(t.structure, tilestructures.tile_daoe_single):
                stiles.append(t)
            elif isinstance(t.structure, tilestructures.tile_daoe_doublehoriz):
                stiles += _ffakesingle(t, gluelist)
                htiles.append(t)
            elif isinstance(t.structure, tilestructures.tile_daoe_doublevert):
                stiles += _ffakesingle(t, gluelist)
                vtiles.append(t)
            else:
                raise NotImplementedError
        self.stiles = _ft_to_fta(stiles)

        for i in range(0, len(self.stiles)):
            x, y = _ffakedouble_n(i, self.stiles, gluelist)
            htiles += x
            vtiles += y

        self.htiles = _ft_to_fta(htiles)
        self.vtiles = _ft_to_fta(vtiles)


RSEL = (2, 3, 0, 1)
FTI = (1, 0, 1, 0)
FTS = (tilestructures.tile_daoe_doublevert(), tilestructures.tile_daoe_doublehoriz(),
       tilestructures.tile_daoe_doublevert(), tilestructures.tile_daoe_doublehoriz())


def _fdg(dir, gs1, gs2):
    if dir == 0:
        return [gs2[0], gs2[1], gs1[1], gs1[2], gs1[3], gs2[3]]
    elif dir == 1:
        return [gs1[0], gs2[0], gs2[1], gs2[2], gs1[2], gs1[3]]
    elif dir == 2:
        return [gs1[0], gs1[1], gs2[1], gs2[2], gs2[3], gs1[3]]
    elif dir == 3:
        return [gs2[0], gs1[0], gs1[1], gs1[2], gs2[2], gs2[3]]
    else:
        raise ValueError(dir)


def _ffakedouble_n(tn, sta, gluelist, outputonly=True, dir4=False, equiv=None):
    if equiv is None:
        equiv = gluelist.blankequiv()
    if not dir4:
        faketiles = ([], [])
        fti = FTI
    else:
        faketiles = ([], [], [], [])
        fti = (0, 1, 2, 3)
    if sta.sfake[tn]:
        ddir = np.flatnonzero(gluelist.tonum['fakedouble'] == sta.glues[tn])[0]
    for dir in range(0, 4):
        if (sta.use[tn, dir] == 1) and outputonly:
            continue
        if outputonly:
            oti = np.nonzero((equiv[gluelist.complement[sta.glues[tn, dir]]] ==
                              equiv[sta.glues[:, RSEL[dir]]]) &
                             (sta.use[:, RSEL[dir]] == 1) & (sta.used))
        else:
            oti = np.nonzero((equiv[gluelist.complement[sta.glues[tn, dir]]] ==
                              equiv[sta.glues[:, RSEL[dir]]]))
        for i in oti[0]:
            faketiles[fti[dir]].append(
                FTile(
                    color=False,
                    used=True,
                    name=sta.name[tn] + '_{}_'.format(dir) + sta.name[i],
                    structure=FTS[dir],
                    glues=np.array(_fdg(dir, sta.glues[tn, :], sta.glues[i])),
                    use=np.array(_fdg(dir, sta.use[tn, :], sta.use[i])),
                    dfake=dir + 1,
                    sfake=(sta.sfake[tn] or sta.sfake[i])))
        if sta.sfake[tn] and (dir == ddir):
            faketiles[fti[dir]].append(
                FTile(
                    color=False,
                    used=True,
                    name=sta.name[tn] + '_{}_'.format(dir) + sta.name[tn+sta.sfake[tn]],
                    structure=FTS[dir],
                    glues=np.array(_fdg(dir, sta.glues[tn, :], sta.glues[tn+sta.sfake[tn]])),
                    use=np.array(_fdg(dir, sta.use[tn, :], sta.use[tn+sta.sfake[tn]])),
                    dfake=dir + 1,
                    sfake=True))
    return faketiles


# THESE SELECTORS ARE 1-INDEXED!!! 0 corresponds to fake double tile bond.
HSEL = ((1, 0, 5, 6), (2, 3, 4, 0))
VSEL = ((1, 2, 0, 6), (0, 3, 4, 5))


def _ffakesingle(ftile, gluelist):
    # FIXME: should be more generalized.  Currently only tau=2
    if isinstance(ftile.structure, tilestructures.tile_daoe_doublehoriz):
        sel = HSEL
    elif isinstance(ftile.structure, tilestructures.tile_daoe_doublevert):
        sel = VSEL
    else:
        raise NotImplementedError

    # Start by making the tiles, then change around the inputs
    fdb = gluelist.tonum['fakedouble']
    glues = [[([fdb] + list(ftile.glues))[x] for x in y] for y in sel]
    fuse = [[([-1] + list(ftile.use))[x] for x in y] for y in sel]
    use = []
    used = []
    names = [ftile.name + '_fakedouble_a', ftile.name + '_fakedouble_b']
    for gu in zip(glues, fuse):
        if sum(gluelist.strength[g] for g, u in zip(*gu) if u == 1) >= TAU:
            use.append(gu[1])
            used.append(True)
        else:
            use.append([-1, -1, -1, -1])
            used.append(False)
    return [
        FTile(
            color=ftile.color,
            use=u,
            glues=g,
            name=n,
            used=ud,
            structure=tilestructures.tile_daoe_single(),
            dfake=0,
            sfake=sfo) for u, g, n, ud, sfo in zip(use, glues, names, used, [1, -1])
    ]


def _ft_to_fta(ftiles):
    return FTilesArray(
        color=np.array([x.color for x in ftiles]),
        name=np.array([x.name for x in ftiles]),
        glues=np.array([x.glues for x in ftiles]),
        use=np.array([x.use for x in ftiles]),
        used=np.array([x.used for x in ftiles]),
        structure=np.array([x.structure.name for x in ftiles]),
        dfake=np.array([x.dfake for x in ftiles]),
        sfake=np.array([x.sfake for x in ftiles]))


class _FastTileSet():
    def __init__(self, tilesystem):
        self.gluelist = FGlueList(tilesystem.allends + [
            End({
                'name': 'hp',
                'type': 'hairpin',
                'strength': 0
            }),
            End({
                'name': 'fakedouble',
                'type': 'fakedouble',
                'strength': 0
            })
        ])
        self.tilelist = FTileList(tilesystem.tiles + sum(
            [x.named_rotations()
             for x in tilesystem.tiles], TileList()), self.gluelist)

    def applyequiv(self, ts, equiv):
        ts = ts.copy()
        alreadythere = []
        for tile in ts.tiles:
            tile.ends = [
                self.gluelist.name[equiv[self.gluelist.tonum[e]]]
                for e in tile.ends
            ]
            if (tile.ends, 'label' in tile.keys()) in alreadythere:
                tile['fake'] = True
                continue
            rs = [tile] + tile.rotations
            alreadythere += [(t.ends, 'label' in t.keys()) for t in rs]
        if 'seed' in ts.keys():
            for t in ts.seed['adapters']:
                if 'ends' in t.keys():
                    t['ends'] = [
                        self.gluelist.name[equiv[self.gluelist.tonum[e]]]
                        for e in t['ends']
                    ]
        ts['info'] = ts.get('info', dict())
        ts['info']['fgluemerge'] = ts['info'].get('fgluemerge', list())
        ts['info']['fgluemerge'].append([int(x) for x in equiv])
        return ts

    def togluemergespec(self, ts, equiv):
        gms = util.GlueMergeSpec()
        for i in range(0, len(equiv)):
            if i != equiv[i]:
                gms.add(self.gluelist.name[equiv[i]], self.gluelist.name[i])
        return gms


def ptins(fts, equiv=None, tau=2):
    """Calculate potential tile attachments to input neighborhoods"""
    ptins = []
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    for ta in [fts.tilelist.stiles, fts.tilelist.htiles, fts.tilelist.vtiles]:
        ptin = []
        for ti in np.arange(0, len(ta.used))[ta.used]:
            # Only iterate through used tiles
            isel = ta.use[ti] == 1  # used edges
            gsel = equiv[ta.glues[ti, isel]] == equiv[ta.glues[:, isel]]
            # matching glues
            strs = np.sum(
                fts.gluelist.strength[ta.glues[ti, isel]] * gsel,
                axis=1)  # strengths of matching
            matches = np.nonzero((strs >= tau) & (ta.dfake == 0))
            # indices of matching
            # (excludes fake doubles, which can't actually attach
            # (because they are
            # actually two singles) to their own local neighborhoods)
            ptin.append(matches)
        ptins.append(ptin)
    return ptins


def is_2go_single_nn(fts,
                     tn,
                     un,
                     equiv,
                     tau=2,
                     in2go=None,
                     retall=False,
                     also22go=False):
    # Before starting, check to see if t and u are actually the same:
    if np.all(equiv[fts.tilelist.stiles.glues[fts.tilelist.stiles.used][tn]] ==
              equiv[fts.tilelist.stiles.glues[un]]):
        if not also22go:
            if not retall:
                return False
            else:
                return False, None
        else:
            if not retall:
                return False, False
            else:
                return False, False, None, None

    if in2go is None:
        in2go = _ffakedouble_n(
            np.flatnonzero(fts.tilelist.stiles.used)[tn],
            fts.tilelist.stiles,
            fts.gluelist,
            outputonly=True,
            dir4=True)
    incorrect = _ffakedouble_n(
        un,
        fts.tilelist.stiles,
        fts.gluelist,
        outputonly=False,
        dir4=True,
        equiv=equiv)
    is2go = None
    for tvs, uws in zip(in2go, incorrect):
        for tv in tvs:
            for uw in uws:
                if np.sum(fts.gluelist.strength[tv.glues[(
                        equiv[tv.glues] == equiv[uw.glues]) & (
                            tv.use == 1)]]) >= tau:
                    if also22go and uw.sfake:
                        if not retall:
                            return True, True
                        else:
                            return True, True, (ts, uw), None
                    elif also22go:
                        is2go = (tv, uw)
                        for dirr, (tvs2,
                                   uws2) in enumerate(zip(in2go, incorrect)):
                            if dirr + 1 == tv.dfake:
                                continue
                            for tv2 in tvs2:
                                for uw2 in uws2:
                                    if np.sum(fts.gluelist.strength[tv2.glues[(
                                            equiv[tv2.glues] == equiv[uw2.glues]
                                    ) & (tv2.use == 1)]]) >= tau:
                                        if not retall:
                                            return True, True
                                        else:
                                            return True, True, (tv, uw), (tv2,
                                                                          uw2)
                    elif not retall:
                        return True
                    else:
                        return True, (tv, uw)
    if also22go:
        if retall:
            return (not (is2go is None)), False, is2go, None
        else:
            return (not (is2go is None)), False
    elif not retall:
        return False
    else:
        return False, None


def gen_2go_single_ins(fts, tau=2):
    ssel = np.flatnonzero(fts.tilelist.stiles.used)
    ins = []
    for tn in ssel:
        ins.append(
            _ffakedouble_n(
                tn,
                fts.tilelist.stiles,
                fts.gluelist,
                outputonly=True,
                dir4=True))
    return ins


def gen_2go_profile(fts, equiv=None, ins2go=None, also22go=False):
    if ins2go is None:
        ins2go = gen_2go_single_ins(fts, tau=2)
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    sens1s = ptins(fts, tau=1, equiv=equiv)[0]
    sens2s = []
    sens22s = []
    for tn, (uns, in2go) in enumerate(zip(sens1s, ins2go)):
        x2 = []
        x22 = []
        for un in uns[0]:
            if also22go:
                s2, s22 = is_2go_single_nn(
                    fts, tn, un, equiv, tau=2, in2go=in2go, also22go=True)
            else:
                s2 = is_2go_single_nn(
                    fts, tn, un, equiv, tau=2, in2go=in2go, also22go=False)
            if s2:
                x2.append(un)
            if also22go and s22:
                x22.append(un)
        sens2s.append(x2)
        if also22go:
            sens22s.append(x22)

    if also22go:
        return sens2s, sens22s
    else:
        return sens2s


def is_2go_equiv(fts, equiv=None, ins2go=None, origsens=None):
    if ins2go is None:
        ins2go = gen_2go_single_ins(fts, tau=2)
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    if origsens is None:
        origsens = gen_2go_profile(fts)
    sens1s = ptins(fts, tau=1, equiv=equiv)[0]

    for tn, (uns, in2go, os) in enumerate(zip(sens1s, ins2go, origsens)):
        for un in uns[0]:
            r, p = is_2go_single_nn(
                fts, tn, un, equiv, tau=2, in2go=in2go, retall=True)
            if (r and (un not in os)):
                # Next if checks to see if sensitive tile is actually identical
                # to an already sensitive tile:
                # FIXME: this would not be necessary if we stored whether a
                # tile was a duplicate.  However, that would break the
                # glue-equiv-only method...
                if not np.any(
                        np.all(
                            equiv[fts.tilelist.stiles.glues[os]] ==
                            equiv[fts.tilelist.stiles.glues[un]],
                            axis=1)):
                    return False, (
                        fts.tilelist.stiles.name[fts.tilelist.stiles.used][tn],
                        fts.tilelist.stiles.name[un]), p
    return True, None, None


def is_22go_equiv(fts, equiv=None, ins2go=None, orig22go=None):
    if ins2go is None:
        ins2go = gen_2go_single_ins(fts, tau=2)
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    if orig22go is None:
        _, orig22go = gen_2go_profile(fts, also22go=True)
    sens1s = ptins(fts, tau=1, equiv=equiv)[0]

    for tn, (uns, in2go, os22) in enumerate(zip(sens1s, ins2go, orig22go)):
        for un in uns[0]:
            r2, r22, p2, p22 = is_2go_single_nn(
                fts,
                tn,
                un,
                equiv,
                tau=2,
                in2go=in2go,
                also22go=True,
                retall=True)
            if (r22 and (un not in os22)):
                # Next if checks to see if sensitive tile is actually identical
                # to an already sensitive tile:
                # FIXME: this would not be necessary if we stored whether a
                # tile was a duplicate.  However, that would break the
                # glue-equiv-only method...
                if not np.any(
                        np.all(
                            equiv[fts.tilelist.stiles.glues[os22]] ==
                            equiv[fts.tilelist.stiles.glues[un]],
                            axis=1)):
                    return False, (
                        fts.tilelist.stiles.name[fts.tilelist.stiles.used][tn],
                        fts.tilelist.stiles.name[un]), p22
    return True, None, None


def fta_to_ft(stiles, un, sel=None):
    if sel is None:
        sel = np.ones_like(stiles.used, dtype=bool)
    return FTile(
        color=stiles.color[sel][un],
        use=stiles.use[sel, :][un, :],
        glues=stiles.glues[sel, :][un, :],
        name=stiles.name[sel][un],
        used=stiles.used[sel][un],
        structure=stiles.structure[sel][un],
        dfake=stiles.dfake[sel][un],
        sfake=stiles.sfake[sel][un])


def isatamequiv(fts, equiv, initptins=None):
    if initptins is None:
        initptins = ptins(fts, fts.gluelist.blankequiv())
    npt = ptins(fts, equiv)
    for x, y, tl in zip(initptins, npt, [
            fts.tilelist.stiles, fts.tilelist.htiles, fts.tilelist.vtiles
    ]):
        for xx, yy in zip(x, y):
            if len(xx[0]) == len(yy[0]) and (np.all(
                    xx[0] == yy[0])):  # No change
                continue
            elif len(xx[0]) == 1:  # Deterministic start
                mm = ~np.all(
                    (equiv[tl.glues[xx]] == equiv[tl.glues[yy]]),
                    axis=1) | ~(tl.color[xx] == tl.color[yy])
                if np.any(mm):
                    return False, (tl.name[xx[0][0]], tl.name[yy[0][mm][0]])
            elif len(xx[0]) == 0:
                return False, None
            else:
                raise NotImplementedError(xx)
    return True, None


def tilemerge(fts, equiv, t1, t2, preserveuse=False):
    if t1.structure.name != t2.structure.name:
        raise ValueError
    if t1.color != t2.color:
        raise ValueError
    for g1, g2 in zip(t1.glues, t2.glues):
        equiv = fts.gluelist.domerge(equiv, g1, g2, preserveuse=preserveuse)
    return equiv


def _findpotentialtilemerges(fts, equiv):
    ppairs = []
    for ti in range(0, len(fts.tilelist.tiles)):
        t1 = fts.tilelist.tiles[ti]
        if not t1.used:
            continue
        ppairs += [
            (t1, t) for t in fts.tilelist.tiles[ti:]
            if (t1.color == t.color) and (
                t1.structure.name == t.structure.name) and (t1.name != t.name)
        ]
    shuffle(ppairs)
    return ppairs


def _findpotentialgluemerges(fts, equiv):
    ppairs = []
    for g1 in np.arange(0, len(fts.gluelist.strength)):
        g2s = g1 + 1 + np.nonzero(
            (fts.gluelist.strength[g1 + 1:] == fts.gluelist.strength[g1]) &
            (fts.gluelist.structure[g1 + 1:] == fts.gluelist.structure[g1]))[0]
        ppairs += [(g1, g2) for g2 in g2s]
    shuffle(ppairs)
    return ppairs


def _recfix(fts,
            equiv,
            tp,
            initptins,
            check2go=False,
            ins2go=None,
            orig2go=None,
            orig22go=None,
            check22go=False,
            checkld=False,
            chain=None,
            preserveuse=False):
    log = logging.getLogger(__name__)
    if chain is None:
        chain = []
    equiv = tilemerge(
        fts,
        equiv,
        fts.tilelist.totile[tp[0]],
        fts.tilelist.totile[tp[1]],
        preserveuse=preserveuse)
    ae, badpair = isatamequiv(fts, equiv, initptins=initptins)
    if check2go and ae:
        ae, badpair, _ = is_2go_equiv(
            fts, equiv, ins2go=ins2go, origsens=orig2go)
    if ae and check22go:
        ae, badpair, _ = is_22go_equiv(
            fts, equiv, ins2go=ins2go, orig22go=orig22go)
    if ae and checkld:
        ld_e = fld.latticedefects(fts, 'e', equiv=equiv)
        ld_w = fld.latticedefects(fts, 'w', equiv=equiv)
        if (len(ld_e) > 0) or (len(ld_w) > 0):
            raise ValueError
    if ae:
        log.debug(
            "Recfix succeeds with {}, {}, {}".format(tp[0], tp[1], chain))
        return equiv
    elif badpair is None:
        raise ValueError
    else:
        chain.append(tp)
        return _recfix(
            fts,
            equiv,
            badpair,
            initptins,
            check2go,
            ins2go,
            orig2go,
            orig22go=orig22go,
            check22go=check22go,
            checkld=checkld,
            chain=chain,
            preserveuse=preserveuse)


def _tilereduce(fts,
                equiv=None,
                check2go=False,
                initptins=None,
                check22go=False,
                checkld=False,
                preserveuse=False):
    log = logging.getLogger(__name__)
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    todo = _findpotentialtilemerges(fts, equiv)
    if initptins is None:
        initptins = ptins(fts)
    if check2go or check22go:
        ins2go = gen_2go_single_ins(fts)
        origsens, orig22go = gen_2go_profile(fts, ins2go=ins2go, also22go=True)
    else:
        ins2go = None
        orig22go = None
        origsens = None
    for todoi, (t1, t2) in enumerate(todo):
        try:
            nequiv = tilemerge(fts, equiv, t1, t2, preserveuse=preserveuse)
        except ValueError:
            continue
        ae, badpair = isatamequiv(fts, nequiv, initptins=initptins)
        if ae and check2go:
            ae, badpair, _ = is_2go_equiv(
                fts, nequiv, ins2go=ins2go, origsens=origsens)
        if ae and check22go:
            ae, badpair, _ = is_22go_equiv(
                fts, nequiv, ins2go=ins2go, orig22go=orig22go)
        if ae and checkld:
            ld_e = fld.latticedefects(fts, 'e', equiv=nequiv)
            ld_w = fld.latticedefects(fts, 'w', equiv=nequiv)
            if (len(ld_e) > 0) or (len(ld_w) > 0):
                continue
        if ae:
            equiv = nequiv
            log.debug("Reduced {}, {} ({} of {} pairs done)".format(
                t1.name, t2.name, todoi, len(todo)))
        elif badpair is None:
            continue
        else:
            try:
                equiv = _recfix(
                    fts,
                    equiv,
                    badpair,
                    initptins,
                    check2go,
                    ins2go,
                    origsens,
                    check22go=check22go,
                    orig22go=orig22go,
                    checkld=checkld,
                    chain=[(t1.name, t2.name)],
                    preserveuse=preserveuse)
            except ValueError:
                continue
            except KeyError:
                continue
    return equiv


def _gluereduce(fts,
                equiv=None,
                check2go=False,
                check22go=False,
                checkld=False,
                initptins=None,
                preserveuse=False):
    log = logging.getLogger(__name__)
    if equiv is None:
        equiv = fts.gluelist.blankequiv()
    todo = _findpotentialgluemerges(fts, equiv)
    if initptins is None:
        initptins = ptins(fts)
    if check2go or check22go:
        ins2go = gen_2go_single_ins(fts)
        origsens, orig22go = gen_2go_profile(fts, ins2go=ins2go, also22go=True)
    else:
        ins2go = None
        origsens = None
        orig22go = None
    for todoi, (g1, g2) in enumerate(todo):
        try:
            nequiv = fts.gluelist.domerge(
                equiv, g1, g2, preserveuse=preserveuse)
        except ValueError:
            continue
        ae, badpair = isatamequiv(fts, nequiv, initptins=initptins)
        if ae and check2go:
            ae, badpair, _ = is_2go_equiv(
                fts, nequiv, ins2go=ins2go, origsens=origsens)
        if ae and check22go:
            ae, badpair, _ = is_22go_equiv(
                fts, nequiv, ins2go=ins2go, orig22go=orig22go)
        if ae and checkld:
            ld_e = fld.latticedefects(fts, 'e', equiv=nequiv)
            ld_w = fld.latticedefects(fts, 'w', equiv=nequiv)
            if (len(ld_e) > 0) or (len(ld_w) > 0):
                continue
        if ae:
            equiv = nequiv
            log.debug("Glue reduction: {}, {} ({}/{} done)".format(
                fts.gluelist.name[g1], fts.gluelist.name[g2], todoi, len(
                    todo)))
        elif badpair is None:
            continue
        else:
            try:
                equiv = _recfix(
                    fts,
                    equiv,
                    badpair,
                    initptins,
                    check2go,
                    ins2go,
                    origsens,
                    check22go=check22go,
                    checkld=checkld,
                    chain=[(fts.gluelist.name[g1], fts.gluelist.name[g2])],
                    preserveuse=preserveuse)
            except ValueError:
                continue
            except KeyError:
                continue
    return equiv

def _single_reduce_tiles(p):    
    feq, ts, fts, params = p
    #starttime = time.time()
    initptins = ptins(fts)
    e = _tilereduce(fts, equiv=feq, initptins=initptins, **params)
    #endtime = time.time()
    #te = fts.applyequiv(ts, e)
    #nt = len([y for y in te.tiles if 'fake' not in y.keys()])
    #ng = len(te.allends)

    #a = isatamequiv(fts, e)
    #g2 = is_2go_equiv(fts, e)[0]
    #g22 = is_22go_equiv(fts, e)[0]
    #lde = len(fld.latticedefects(fts, equiv=e, direction='e')) == 0
    #ldw = len(fld.latticedefects(fts, equiv=e, direction='w')) == 0
    #print('TR: Found {}t, {}g: aTAM {} 2GO {} 22GO {} LD {}/{} in {}s'.format(
    #    nt, ng, a, g2, g22, lde, ldw, endtime - starttime))
    #te.to_file(nb + 'step1-{}t{}g.yaml'.format(nt, ng))
    return e

def _single_reduce_ends(p):
    equiv, ts, fts, params = p
    #starttime = time.time()
    initptins = ptins(fts)
    e = _gluereduce(fts, equiv=equiv, initptins=initptins, **params)
    # FIXME: all this is code for checking things and printing logs.
    #endtime = time.time()
    #te = fts.applyequiv(ts, e)
    #nt = len([y for y in te.tiles if 'fake' not in y.keys()])
    #ng = len(te.allends)
    #a = isatamequiv(fts, e)
    #g2 = is_2go_equiv(fts, e)[0]
    #g22 = is_22go_equiv(fts, e)[0]
    #lde = len(fld.latticedefects(fts, equiv=e, direction='e')) == 0
    #ldw = len(fld.latticedefects(fts, equiv=e, direction='w')) == 0
    #print('ER: Found {}t, {}g: aTAM {} 2GO {} 22GO {} LD {}/{} in {}s'.format(
    #    nt, ng, a, g2, g22, lde, ldw, endtime - starttime))
    #te.to_file(nb + '-step2-{}t{}g.yaml'.format(nt, ng))
    return e

def reduce_tiles(tileset, preserve=('s22','ld'), tries=10, threads=1, returntype='equiv', best=1, key=None, initequiv=None):
    """
    Apply tile reduction algorithm, preserving some set of properties, and using a multiprocessing pool.

    Parameters
    ----------
    tileset: TileSet  
        The system to reduce. 

    preserve: a tuple or list of strings, optional
        The properties to preserve.  Currently supported are 's1' for first order
        sensitivity, 's2' for second order sensitivity, 's22' for two-by-two sensitivity,
        'ld' for small lattice defects, and 'gs' for glue sense (to avoid spurious
        hierarchical attachment).  Default is currently ('s22', 'ld').

    tries: int, optional
        The number of times to run the algorithm.

    threads: int, optional
        The number of threads to use (using multiprocessing).

    returntype: 'TileSet' or 'equiv' (default 'equiv')
        The type of object to return.  If 'equiv', returns an array of glue equivalences
        (or list, if best != 1) that can be applied to the tileset with apply_equiv, or used 
        for further reduction.  If 'TileSet', return a TileSet with the equiv already applied
        (or a list, if best != 1).

    best: int or None, optional
        The number of systems to return.  If 1, the result will be returned
        directly; if k > 1, a list will be returned of the best k results (per cmp);
        if k = None, a list of *all* results will be returned, sorted by cmp. (default 1)

    key: function (ts, equiv1, equiv2) -> some number/comparable
        A comparison function for equivs, to sort the results. FIXME: documentation needed.
        Default (if None) here is to sort by number of glues in the system, regardless of number 
        of tiles.

    initequiv: equiv
        If provided, the equivalence array to start from.  If None, start from the tileset without
        any merged glues.

    Returns
    -------
    reduced: single TileSet
        The reduced system/systems
    """


    fts = _FastTileSet(tileset)

    if key is None:
        key = lambda x: len(np.unique(x)) # number of unique numbers in equiv, equivalent
                                          # to number of glues in reduced system. FIXME?

    #FIXME: could do a better job here
    params = {
        'check2go': 's2' in preserve,
        'check22go': 's22' in preserve,
        'checkld': 'ld' in preserve,
        'preserveuse': 'gs' in preserve
        }

    if initequiv is None:
        initequiv = fts.gluelist.blankequiv()

    if threads > 1:
        from multiprocessing import Pool
        with Pool(threads) as pool:
            equivs = pool.map(_single_reduce_tiles, [[initequiv, tileset, fts, params]] * tries)
    else:
        equivs = [_single_reduce_tiles(x) for x in ([[initequiv, tileset, fts, params]] * tries)]

    equivs.sort(key=key)

    if returntype == 'TileSet':
        equivs = [ tileset.apply_equiv(equiv) for equiv in equivs ]

    if best == 1:
        return equivs[0]
    else:
        return equivs[0:best]


def reduce_ends(tileset, preserve=('s22','ld'), tries=10, threads=1, returntype='equiv', best=1, key=None, initequiv=None):
    """
    Apply end reduction algorithm, preserving some set of properties, and using a multiprocessing pool.

    Parameters
    ----------
    tileset: TileSet  
        The system to reduce. 

    preserve: a tuple or list of strings, optional
        The properties to preserve.  Currently supported are 's1' for first order
        sensitivity, 's2' for second order sensitivity, 's22' for two-by-two sensitivity,
        'ld' for small lattice defects, and 'gs' for glue sense (to avoid spurious
        hierarchical attachment).  Default is currently ('s22', 'ld').

    tries: int, optional
        The number of times to run the algorithm.

    threads: int, optional
        The number of threads to use (using multiprocessing).

    returntype: 'TileSet' or 'equiv' (default 'equiv')
        The type of object to return.  If 'equiv', returns an array of glue equivalences
        (or list, if best != 1) that can be applied to the tileset with apply_equiv, or used 
        for further reduction.  If 'TileSet', return a TileSet with the equiv already applied
        (or a list, if best != 1).

    best: int or None, optional
        The number of systems to return.  If 1, the result will be returned
        directly; if k > 1, a list will be returned of the best k results (per cmp);
        if k = None, a list of *all* results will be returned, sorted by cmp. (default 1)

    key: function (ts, equiv1, equiv2) -> some number/comparable
        A comparison function for equivs, to sort the results. FIXME: documentation needed.
        Default (if None) here is to sort by number of glues in the system, regardless of number 
        of tiles.

    initequiv: equiv
        If provided, the equivalence array to start from.  If None, start from the tileset without
        any merged glues.

    Returns
    -------
    reduced: single TileSet
        The reduced system/systems
    """


    fts = _FastTileSet(tileset)

    if key is None:
        key = lambda x: len(np.unique(x)) # number of unique numbers in equiv, equivalent
                                          # to number of glues in reduced system. FIXME?

    #FIXME: could do a better job here
    params = {
        'check2go': 's2' in preserve,
        'check22go': 's22' in preserve,
        'checkld': 'ld' in preserve,
        'preserveuse': 'gs' in preserve
        }

    if initequiv is None:
        initequiv = fts.gluelist.blankequiv()

    if threads > 1:
        from multiprocessing import Pool
        with Pool(threads) as pool:
            equivs = pool.map(_single_reduce_ends, [[initequiv, tileset, fts, params]] * tries)
    else:
        equivs = [_single_reduce_ends(x) for x in ([[initequiv, tileset, fts, params]] * tries)]

    equivs.sort(key=key)

    if returntype == 'TileSet':
        equivs = [ tileset.apply_equiv(equiv) for equiv in equivs ]

    if best == 1:
        return equivs[0]
    else:
        return equivs[0:best]

    



# End reordering code to optimize placement.

from . import anneal
from . import sensitivity as sens
import stickydesign as sd
import numpy.random as random
import numpy as np
import random as pyrand
from copy import deepcopy
import numpy.ma as ma

def flatten(seq):
    for item in seq:
        if is_string_like(item) or not iterable(item):
            yield item
        else:
            for subitem in flatten(item):
                yield subitem

def ecomp(x):
    if x[-1]=='/':
        return x[:-1]
    else:
        return x+'/'

class FseqState:
    def __init__(self, seqs=None):
        if not seqs:
            self.seqs = {}
        else:
            self.seqs = seqs
    def copy(self):
        return FseqState({'DT': self.seqs['DT'].copy(), 'TD': self.seqs['TD'].copy()})

class cachedarray:
    def __init__(self, func, shape):
        self.arr = ma.masked_all(shape)
        self.func = func
    def __getitem__(self, index):
        if self.arr.mask[index]:
            self.arr[index] = self.func(*index)
        return self.arr[index]

from copy import deepcopy
class FastState:
    def __init__(self, state):
        self.state = state
    def copy(self):
        return deepcopy(self)
    def __getitem__(self, it):
        return self.state[it]


class EndSystemFseq:
    def __init__(self, tilesys, newends=None, pairs=None, energetics=None, inputpairs=False, multiscore=False):
        tilesys = deepcopy(tilesys)
        self.ends = tilesys.ends
        self.tiles = tilesys.tiles
        self.tilesystem = tilesys
        self.ef = energetics
        self.multiscore = multiscore
        if not pairs:
            pairs = sens.consolidate_pairs( sens.senspairs(tilesys), comcomp=1, onlytop=True )

        if inputpairs:
            inputpairs = [  tuple(z[:-1] for z,i in zip(x['ends'],x['input']) if i)
                             for x in self.tiles ]
            print(inputpairs)
        self.inputpairs = inputpairs
            
        self.names = {}
        fseqsTD, self.names['TD'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'TD']))
        fseqsDT, self.names['DT'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'DT']))
        self.seqs = {}
        self.seqs['TD'] = sd.endarray(fseqsTD,'TD')
        self.seqs['DT'] = sd.endarray(fseqsDT,'DT')
        self.initstate = FastState({'DT': np.arange(0,len(self.seqs['DT'])), 'TD': np.arange(0,len(self.seqs['TD']))})
        self.enlocs = {}
        for i, endn in enumerate(self.names['TD']):
            self.enlocs[endn] = (i,'TD')
        for i, endn in enumerate(self.names['DT']):
            self.enlocs[endn] = (i,'DT')        

        # ends that can be reordered.
        if newends:
            self.mutableTD = [ i for i,t in [self.enlocs[x] for x in newends] if t=='TD']
            self.mutableDT = [ i for i,t in [self.enlocs[x] for x in newends] if t=='DT']
        else:
            self.mutableTD = range(0,len(fseqsTD))
            self.mutableDT = range(0,len(fseqsDT))


        self.pairdict = {}
        for pairclass,memberset in pairs.items():
            for x,y in memberset:
                self.pairdict[(x,ecomp(y))] = pairclass
                self.pairdict[(y,ecomp(x))] = pairclass
        tdsh = ( len(self.seqs['TD']), len(self.seqs['TD']) )
        dtsh = ( len(self.seqs['DT']), len(self.seqs['DT']) )
            
        if not multiscore:
            # Get the mean non-spurious interaction
            self.meangse = 0.5*( np.mean(self.ef.matching_uniform( self.seqs['TD'] ))+np.mean(self.ef.matching_uniform( self.seqs['DT'] )) )

            self.mult = {'1NGO': np.exp(-2.0*self.meangse), '2NGO': np.exp(-1.65*self.meangse), '1GO': np.exp(-1.5*self.meangse),
                         '2GO': np.exp(-1.1*self.meangse), 'I': np.exp(-1.0*self.meangse)}

            self.ecache_cc = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].comps) , tdsh ),
                                'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].comps) , dtsh ) }
            self.ecache_ce = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].ends) , tdsh ),
                                'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].ends) , dtsh ) }
            self.ecache_ec = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].comps) , tdsh ),
                                'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].comps) , dtsh ) }
            self.ecache_ee = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].ends) , tdsh ),
                                'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].ends) , dtsh ) }
        else:
            self.ecache_cc = []
            self.ecache_ee = []
            self.ecache_ce = []
            self.ecache_ec = []
            self.mult = []
            self.meangse = []
            for ef in self.ef:
                self.meangse.append(0.5*( np.mean(ef.matching_uniform( self.seqs['TD'] ))+np.mean(ef.matching_uniform( self.seqs['DT'] )) ))

                self.mult.append({'1NGO': np.exp(-2.0*self.meangse[-1]), '2NGO': np.exp(-1.65*self.meangse[-1]), '1GO': np.exp(-1.5*self.meangse[-1]),
                             '2GO': np.exp(-1.1*self.meangse[-1]), 'I': np.exp(-1.0*self.meangse[-1])})

                self.ecache_cc.append({  'TD': cachedarray( lambda x,y: ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].comps) , tdsh ),
                                    'DT': cachedarray( lambda x,y: ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].comps) , dtsh ) })
                self.ecache_ce.append({  'TD': cachedarray( lambda x,y: ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].ends) , tdsh ),
                                    'DT': cachedarray( lambda x,y: ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].ends) , dtsh ) })
                self.ecache_ec.append({  'TD': cachedarray( lambda x,y: ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].comps) , tdsh ),
                                    'DT': cachedarray( lambda x,y: ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].comps) , dtsh ) })
                self.ecache_ee.append({  'TD': cachedarray( lambda x,y: ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].ends) , tdsh ),
                                    'DT': cachedarray( lambda x,y: ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].ends) , dtsh ) })
                
            
                    
    def slowseqs(self, state):
        "Give the state as the slow version would have"
        return {'DT': self.seqs['DT'][state['DT']], 'TD': self.seqs['TD'][state['TD']] }

    def mutate(self, state):
        # Start by deciding to swap TD or DT ends.
        if random.rand() > 1.0*len(self.mutableTD)/(len(self.mutableDT)+
                                                    len(self.mutableTD)):
            a = random.choice(self.mutableDT)
            b = random.choice(self.mutableDT)
            state['DT'][[a,b]] = state['DT'][[b,a]]
        else:
            a = random.choice(self.mutableTD)
            b = random.choice(self.mutableTD)
            state['TD'][[a,b]] = state['TD'][[b,a]]

    def score(self, state):
        
        sc = 0.0
        if self.inputpairs:
            if self.multiscore:
                ip = np.zeros((len(self.ef),len(self.inputpairs)))
            else:
                ip = np.zeros(len(self.inputpairs))
            for i, (xn, yn) in enumerate(self.inputpairs):
                if xn[-1] == '/':
                    xn = xn[:-1]
                if yn[-1] == '/':
                    yn = yn[:-1]
                xi,xt = self.enlocs[xn]
                yi,yt = self.enlocs[yn]
                if self.multiscore:
                    scc = 0
                    for j in range(len(self.ef)):
                        ip[j][i] = np.abs((self.ecache_ec[j][xt][state[xt][xi], state[xt][xi]] +
                                           self.ecache_ec[j][yt][state[yt][yi], state[yt][yi]]))
                else:
                    ip[i] = np.abs((self.ecache_ec[xt][state[xt][xi], state[xt][xi]] +
                                self.ecache_ec[yt][state[yt][yi], state[yt][yi]]))
            if self.multiscore:
                sc += np.mean(np.array([z['I'] for z in self.mult]) * np.exp(np.ptp(ip, axis=1)))
            else:
                sc += self.mult['I'] * np.exp(np.ptp(ip))
        
        for (xn,yn),pairclass in self.pairdict.items():
            # set comp flags
            xc = False
            yc = False
            if xn[-1] == '/':
                xc = True
                xn = xn[:-1]
            if yn[-1] == '/':
                yc = True
                yn = yn[:-1]
                
            # get end indexes and types
            xi,xt = self.enlocs[xn]
            yi,yt = self.enlocs[yn]
            #print "%s, %s: (%s %s)" % (xn,yn,xt,yt)
            # skip if not same type
            if xt != yt: continue
            if not self.multiscore:
                if yc and xc:
                    val = self.ecache_cc[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].comps)[0]
                elif xc:
                    val = self.ecache_ce[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].ends)[0]
                elif yc:
                    val = self.ecache_ec[xt][state[xt][xi],state[yt][yi]] # state[f.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].comps)[0]
                else:
                    val = self.ecache_ee[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].ends)[0]

                sc += self.mult[pairclass]*np.exp( val )
            else:
                scc = 0
                for i in range(len(self.ef)):
                    if yc and xc:
                        val = self.ecache_cc[i][xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].comps)[0]
                    elif xc:
                        val = self.ecache_ce[i][xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].ends)[0]
                    elif yc:
                        val = self.ecache_ec[i][xt][state[xt][xi],state[yt][yi]] # state[f.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].comps)[0]
                    else:
                        val = self.ecache_ee[i][xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].ends)[0]

                    scc += self.mult[i][pairclass]*np.exp( val )
                sc += scc/len(self.ef)
                
        return sc   

            
wcd = {   'a': 't',
         'b': 'v',
         'c': 'g',
         'd': 'h',
         'g': 'c',
         'h': 'd',
         'k': 'm',
         'm': 'k',
         'n': 'n',
         's': 's',
         't': 'a',
         'v': 'b',
         'w': 'w' }
         
def wc(seqstr):
    return ''.join(wcd[x] for x in reversed(seqstr))
